"""
PySCeS - Python Simulator for Cellular Systems (http://pysces.sourceforge.net)

Copyright (C) 2004-2022 B.G. Olivier, J.M. Rohwer, J.-H.S Hofmeyr all rights reserved,

Brett G. Olivier (bgoli@users.sourceforge.net)
Triple-J Group for Molecular Cell Physiology
Stellenbosch University, South Africa.

Permission to use, modify, and distribute this software is given under the
terms of the PySceS (BSD style) license. See LICENSE.txt that came with
this distribution for specifics.

NO WARRANTY IS EXPRESSED OR IMPLIED.  USE AT YOUR OWN RISK.
Brett G. Olivier
"""

from __future__ import division, print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from .version import __version__

__doc__ = '''SBML reading/writing module - now replaced by PySCeS Core2'''

import os, sys
from time import sleep, strftime
from getpass import getuser

try:
    input = raw_input  # Py2 compatibility
except NameError:
    pass

if sys.platform == 'win32':
    try:
        import libsbml as SBML
    except Exception as e:
        print('Windows sbml load error', e)
else:
    try:
        import libsbml as SBML
    except Exception as e:
        print('Posix sbml load error', e)


class PyscesSBML:
    """The PySCeS interface to libSBML and SBML utilities"""

    mode_number_format = '%2.4e'
    sbml_level = 2
    _debug = 0
    SBML = SBML

    def SBML_buildBasicModel(
        self,
        mod,
        filename,
        slvl=2,
        dir=None,
        substance=(1, 0),
        volume=(1, 0),
        time=(1, 0),
        arules=None,
        notes=None,
    ):
        """
        SBML_buildBasicModel(mod,filename,slvl=2,dir=None)

        Create a basic SBML model.
        Arguments:
        =========
        mod: an active PySCeS model object
        filename: the output SBML file name
        slvl [default=2]: the SBML level that should be used
        dir [default=None]: the output directory
        substance [default=(1,0)]: the model substance unit - SBML default is "mole"
        volume [default=(1,0)]: the model volume unit - SBML default is "litre"
        time [default=(1,0)]: the model time unit - SBML default is "second"

        """
        self.SBML_createModel(mod, filename, slvl, dir)
        self.SBML_setCompartment()
        self.SBML_setNotes(txt=notes)
        self.SBML_setUnits(substance=substance, volume=volume, time=time)
        if arules != None:
            self.SBML_setAssignmentRules(arules)
        self.SBML_setSpecies()
        self.SBML_setReactions()
        self.SBML_setModel()

    def write(
        self,
        mod,
        filename,
        slvl=2,
        dir=None,
        substance=(1, 0),
        volume=(1, 0),
        time=(1, 0),
        arules=None,
        notes=None,
    ):
        """
        write(mod,filename,slvl=2,dir=None)

        Write a PySCeS model as an SBML file.

        Arguments:
        =========
        mod: an active PySCeS model object
        filename: the output SBML file name
        slvl [default=2]: the SBML level that should be used
        dir [default=None]: the output directory
        substance [default=(1,0)]: the model substance unit - SBML default is "mole"
        volume [default=(1,0)]: the model volume unit - SBML default is "litre"
        time [default=(1,0)]: the model time unit - SBML default is "second"

        """
        self.SBML_buildBasicModel(
            mod, filename, slvl, dir, substance, volume, time, arules, notes
        )
        self.SBML_writeFile()

    def getSBML_document(
        self, mod, substance=(1, 0), volume=(1, 0), time=(1, 0), arules=None, notes=None
    ):
        """
        Returns an SBML document object

        Arguments:
        =========
        mod: an active PySCeS model object
        substance [default=(1,0)]: the model substance unit - SBML default is "mole"
        volume [default=(1,0)]: the model volume unit - SBML default is "litre"
        time [default=(1,0)]: the model time unit - SBML default is "second"

        """

        filename = 'tempXML'
        slvl = 2
        dir = None
        self.SBML_buildBasicModel(
            mod, filename, slvl, dir, substance, volume, time, arules, notes
        )
        return self.sbml_document

    def getSBML_string(
        self, mod, substance=(1, 0), volume=(1, 0), time=(1, 0), arules=None, notes=None
    ):
        """
        Returns an SBML file as a string

        Arguments:
        =========
        mod: an active PySCeS model object
        substance [default=(1,0)]: the model substance unit - SBML default is "mole"
        volume [default=(1,0)]: the model volume unit - SBML default is "litre"
        time [default=(1,0)]: the model time unit - SBML default is "second"
        """

        filename = 'tempXML'
        slvl = 2
        dir = None
        self.SBML_buildBasicModel(
            mod, filename, slvl, dir, substance, volume, time, arules, notes
        )
        return self.sbml_document.toSBML()

    def __cleanString__(self, s):
        s = s.lstrip()
        s = s.rstrip()
        return s

    def parseForcingFunctions(self):
        self.__forcing_function_dic__ = {}
        ff = self.model_obj._Function_forced.split('\n')
        for f in ff:
            if f != '':
                f = f.split('=')
                f[0] = f[0].replace('self.', '')
                f[1] = f[1].replace('self.', '')
                self.__forcing_function_dic__.setdefault(
                    self.__cleanString__(f[0]), self.__cleanString__(f[1])
                )

    def getAssignmentRules(self):
        self.parseForcingFunctions()
        out = []
        for key in list(self.__forcing_function_dic__.keys()):
            out.append((key, self.__forcing_function_dic__[key]))
        return out

    def SBML_createModel(self, mod, filename, slvl=2, dir=None):
        """
        SBML_createModel(mod,filename,slvl=2,dir=None)

        Set up an SBML document and extract the model NetworkDict

        Arguments:
        =========
        mod: a PySCeS model object
        filename: the output filename
        slvl [default=2]: SBML level required
        dir [default=None]: output directory

        """
        if self._debug:
            print('SBML_createModel')
        self.model_obj = mod
        self.__nDict__ = self.model_obj.__nDict__
        self.model_filename = filename
        if dir == None:
            self.model_dir = self.model_obj.ModelOutput
        else:
            self.model_dir = dir

        self.sbml_level = slvl
        self.sbml_model = self.SBML.Model()
        self.sbml_model.setName(self.model_obj.ModelFile[:-4])
        self.sbml_document = self.SBML.SBMLDocument()
        # new stuff
        self.global_parameters = []
        self.model_compartment_name = None

        # create initdict
        self.__InitStrings__ = [
            s.replace('self.', '') for s in self.model_obj._PysMod__InitStrings
        ]
        self.__InitDict__ = {}
        for ii in self.__InitStrings__:
            l, r = ii.split('=')
            self.__InitDict__.setdefault(
                self.__cleanString__(l), float(self.__cleanString__(r))
            )

        # create forcing function dic
        try:
            self.parseForcingFunctions()
        except:
            print("No pre-defined forcing functions")

        if self.sbml_level == 1:
            if sys.platform == 'win32':
                print(
                    'Due to a bug in self.SBML for Windows writing a lvl 1 file will crash your session writing lvl 2 instead ... sorry'
                )
                self.sbml_document.setLevel(2)
            else:
                self.sbml_document.setLevel(self.sbml_level)
        else:
            self.sbml_document.setLevel(2)

    def SBML_setCompartment(self, name=None, vol=1):
        """
        SBML_setCompartment(name=None,vol=1)

        Initialise SBML compartments (note PySCeS currently utilises a single compartment)

        Arguments:
        =========
        name [default=None]: the compartment name, default is compartment1
        vol [default=1]: the compartment volume

        """
        if self._debug:
            print('SBML_setCompartment')
        comp_def = self.sbml_model.createCompartment()
        if not name:
            self.model_compartment_name = 'Cell'
        else:
            self.model_compartment_name = name
            for char in [' ', '.', '-', '*', '?', '!', '\t', '\n']:
                self.model_compartment_name = self.model_compartment_name.replace(
                    char, '_'
                )
            self.model_compartment_name = name
        comp_def.setId(self.model_compartment_name)
        comp_def.setVolume(vol)

    def SBML_setNotes(self, txt=None):
        notes = '<body xmlns="http://www.w3.org/1999/xhtml">'
        if txt != None:
            notes += '<span style="font-family: Courier New,Courier,monospace;">'
            notes += txt
            notes += '</span>'
        notes += '</body>'
        self.sbml_model.setNotes(notes)

    def SBML_setUnits(self, **kwargs):
        """
        SBML_setUnits(substance=(1,0), volume=(1,0), time=(1,0))

        Set the SBML default units note that the input here is the factor and index multiplying
        the SBML default, so for example the default substance (1,0) is (1*10**0)*mole So e.g.
        if you were specifing default units of millimoles and seconds you would
        set substance=(1,-3) and time=(60,0) i.e. (1*10**-3)*mole and (60*10**0)*seconds

        Arguments:
        =========
        substance [default=(1,0)]: the model substance unit - SBML default is "mole"
        volume [default=(1,0)]: the model volume unit - SBML default is "litre"
        time [default=(1,0)]: the model time unit - SBML default is "second"

        """
        for un in list(kwargs.keys()):
            vdef = self.sbml_model.createUnitDefinition()
            vdef.setId(un)
            vu = self.sbml_model.createUnit()
            if un == 'substance':
                vu.setKind(self.SBML.UnitKind_forName('mole'))
            elif un == 'volume':
                vu.setKind(self.SBML.UnitKind_forName('litre'))
            elif un == 'time':
                vu.setKind(self.SBML.UnitKind_forName('second'))
            vu.setMultiplier(kwargs[un][0])
            vu.setScale(kwargs[un][1])
            vu.setOffset(0)

    def SBML_setSpecies(self):
        """
        SBML_setSpecies()

        Initialise and add species information to the SBML model

        Arguments:
        None

        """
        if self._debug:
            print('SBML_setSpecies')
        reagList = self.model_obj.__species__ + self.model_obj.__fixed_species__
        for reagent in range(len(reagList)):
            s = self.sbml_model.createSpecies()
            s.setId(reagList[reagent])
            s.setName(reagList[reagent])
            s.setCompartment(self.model_compartment_name)
            if reagList[reagent] in self.model_obj.__fixed_species__:
                s.setBoundaryCondition(True)
                s.setConstant(True)
            else:
                s.setBoundaryCondition(False)

            if reagent < len(self.model_obj.__species__):
                reagName = reagList[reagent] + '_init'
            else:
                reagName = reagList[reagent]

            if self.sbml_level == 1:
                s.setInitialAmount(getattr(self.model_obj, reagName))
            else:
                s.setInitialConcentration(getattr(self.model_obj, reagName))

    def SBML_setAssignmentRules(self, rules=[]):

        for rule in rules:
            print(rule)
            self.global_parameters.append(rule[0])
            p = self.sbml_model.createParameter()
            p.setId(rule[0])
            p.setValue(getattr(self.model_obj, rule[0]))
            p.setConstant(False)
            r = self.sbml_model.createAssignmentRule()
            r.setVariable(rule[0])
            r.setFormula(rule[1])
            r.setMathFromFormula()

    def SBML_setReactions(self):
        """
        SBML_setReactions()

        Add kinetic rate laws to the SBMl model

        Arguments:
        None

        """
        if self._debug:
            print('SBML_setReactions')
        # TotSpecies = list(self.model_obj._PysMod__FixedReagents)+list(self.model_obj._PysMod__VarReagents)
        reaction_params = []
        for rxn in self.model_obj._PysMod__ReactionIDs:
            print('Adding reaction:', rxn)
            i = self.sbml_model.createReaction()
            i.setId(rxn)
            ndr = self.model_network_dict[rxn]
            for reagent in ndr['Reagents']:
                stoich = ndr['Reagents'][reagent]
                species = self.SBML.SpeciesReference(
                    reagent.replace('self.', ''), abs(stoich)
                )
                if stoich < 0:
                    i.addReactant(species)
                elif stoich > 0:
                    i.addProduct(species)
                elif stoich == 0:
                    i.addModifier(species)
            # add a volume to convert rate equation to kinetic law
            kineticLaw = ndr['RateEq'].replace('self.', '')
            kineticLaw = kineticLaw.replace('scipy.', '')
            if self.model_compartment_name not in self.model_obj.parameters:
                kineticLaw = self.model_compartment_name + ' * (' + kineticLaw + ')'
            else:
                kineticLaw = kineticLaw
            kineticLaw = self.SBML.KineticLaw(kineticLaw)

            # local parameters retired in favour of globals
            ##  for parameter in ndr['Params']:
            ##  p = parameter.replace('self.','')
            ##  if p not in self.model_obj.__fixed_species__ and p not in self.global_parameters:
            ##  try:
            ##  kineticLaw.addParameter(self.SBML.Parameter(p, getattr(self.model_obj,p)))
            ##  reaction_params.append(p)
            ##  except AttributeError,err :
            ##  print '\n', err
            ##  print "Parameter set error ... are there forcing functions??"
            ##  sleep(0.5)
            i.setKineticLaw(kineticLaw)
            if ndr['Type'] == 'Rever':
                rev = True
            else:
                rev = False
            i.setReversible(rev)

            # Add modifiers to reaction - brett 20050607
            for reac in self.model_obj.__modifiers__:
                if reac[0] == rxn:
                    for x in reac[1]:
                        print(' ' + reac[0] + ' has modifier: ' + x)
                        self.sbml_model.createModifier().setSpecies(x)

        # add extra parameter initialised but not in reactions
        # we have to do this in case the assignment rules are added after we build the model
        hack = list(self.__forcing_function_dic__.keys())

        not_xparams = (
            self.global_parameters
            + reaction_params
            + list(self.model_obj.species)
            + list(self.model_obj.fixed_species)
            + [self.model_compartment_name]
            + hack
        )

        for k in list(self.__InitDict__.keys()):
            if k not in not_xparams:
                print('Adding parameter:', k)
                self.global_parameters.append(k)
                p = self.sbml_model.createParameter()
                p.setId(k)
                p.setValue(getattr(self.model_obj, k))

    def SBML_setModel(self):
        """
        SBML_setModel()

        Add the SBML model to the predefined SBML document

        Arguments:
        None

        """
        if self._debug:
            print('SBML_setModel')
        self.sbml_document.setModel(self.sbml_model)

    def SBML_writeFile(self):
        """
        SBML_writeFile()

        Write the SBML document to predefined output file

        Arguments:
        None

        """

        self.SBML.writeSBML(self.sbml_document, 'pysces_sbml_tmp.xml')
        Fin = open('pysces_sbml_tmp.xml', 'r')
        Fout = open(os.path.join(self.model_dir, self.model_filename + '.xml'), 'w')
        cntr = 0
        try:
            UseR = getuser()
        except:
            UseR = ''
        for line in Fin:
            if cntr == 1:
                Fout.write(
                    '<!-- Created with PySCeS ('
                    + __version__
                    + ') on '
                    + strftime("%a, %d %b %Y %H:%M:%S")
                    + ' by '
                    + UseR
                    + ' -->\n'
                    + line
                )
            else:
                Fout.write(line)
            cntr += 1
        Fout.close()
        Fin.close()

        os.remove('pysces_sbml_tmp.xml')

    def convert2psc(self, filename, dir=None, dirOut=None):
        """
        convert2psc(filename,dir=None,dirOut=None)

        Convert an SBML file into a PySCeS input file

        Arguments:
        =========
        filename: the SBML source file
        dir [default=None]: specify the SBMl file directory
        dirOut [default=None]: the PSC file output directory

        """
        if dir == None:
            dir = os.getcwd()
        File = os.path.join(dir, filename)
        assert os.path.exists(File), "Invalid path"

        self.model_filename = filename

        r = self.SBML.SBMLReader()
        d = r.readSBML(File)
        m = d.getModel()

        def getName(i):
            if d.getLevel() == 1:
                return i.getName()
            else:
                return i.getId()

        reactions = m.getListOfReactions()
        ReactionIDs = []
        for i in reactions:
            ReactionIDs.append(getName(i))

        init_fixed = []
        init_var = []
        init_par = []
        parameters = []

        for i in m.getListOfSpecies():
            parName = getName(i)
            # if a species is a BoundaryCondition or constant it becomes fixed - brett 20050111
            if i.getBoundaryCondition() or i.getConstant():
                if i.getConstant() and not i.getBoundaryCondition():
                    print(
                        parName,
                        ' is set as constant, assuming: BoundaryCondition = True',
                    )
                init_fixed.append((parName, i.getInitialConcentration()))
            else:
                init_var.append((parName, i.getInitialConcentration()))

        NetworkDict = dict(
            [
                (i, dict.fromkeys(['Params', 'RateEq', 'Reagents', 'Type']))
                for i in ReactionIDs
            ]
        )

        for i in reactions:
            rDict = NetworkDict[getName(i)]
            j = i.getKineticLaw()
            par = []
            try:
                for k in j.getListOfParameters():
                    par.append(getName(k))
                    init_par.append((getName(k), k.getValue()))
                    parameters.append(getName(k))
                rDict['Params'] = par
                rDict['RateEq'] = j.getFormula()
                if d.getLevel() == 1:
                    rDict['RateEq'] = rDict['RateEq'].replace(' ', '')
                    rDict['RateEq'] = rDict['RateEq'].replace('^', '**')
            except Exception as err:
                rDict['Params'] = []
                rDict['RateEq'] = ''
                print(err)

            Substrates = []
            Products = []

            for k in i.getListOfReactants():
                species = k.getSpecies()
                stoich = -k.getStoichiometry()
                Substrates.append((species, stoich))

            for k in i.getListOfProducts():
                species = k.getSpecies()
                stoich = k.getStoichiometry()
                Products.append((species, stoich))

            # this is to eliminate zero stoichiometries {0}xyz
            badList = []
            for sub in Substrates:
                if sub[1] == 0:
                    badList.append(sub)
            for bad in badList:
                Substrates.pop(Substrates.index(bad))

            badList = []
            for prod in Products:
                if prod[1] == 0:
                    badList.append(prod)
            for bad in badList:
                Products.pop(Products.index(bad))

            # add source/sink pools to nasty substrate/productless reactions - brett 20050908
            if len(Substrates) == 0:
                Substrates.append(('$pool', -1.0))
            if len(Products) == 0:
                Products.append(('$pool', 1.0))

            # print Substrates
            # print Products

            rDict['Reagents'] = dict(Substrates + Products)
            if i.getReversible() == True:
                t = 'Rever'
            else:
                t = 'Irrev'
            rDict['Type'] = t
            NetworkDict[getName(i)].update(rDict)

        # Add extra model parameters not defined in reactions (apparently)
        if len(m.getListOfParameters()) > 0:
            for x in m.getListOfParameters():
                if getName(x) not in parameters:
                    # print getName(x)
                    init_par.append((getName(x), x.getValue()))

        if dirOut == None:
            self.model_filename = os.path.join(os.getcwd(), self.model_filename)
        else:
            self.model_filename = os.path.join(dirOut, self.model_filename)

        # print 'init_par'
        # print init_par
        # print 'init_var'
        # print init_var
        # print 'init_fixed'
        # print init_fixed

        # sometimes things just work lekker (replaced all the old showS^&t) - brett 20050913
        outFile = open(self.model_filename + '.psc', 'w')
        self.PSC_writeHeader(outFile)
        self.PSC_writeFixedSpeciesList(outFile, init_fixed)
        self.PSC_writeRateEquations(outFile, NetworkDict, number_format='%2.3f')
        self.PSC_writeSpecies(outFile, init_var)
        self.PSC_writeFixedSpecies(outFile, init_fixed)
        self.PSC_writeParameters(outFile, init_par)
        outFile.close()

        # Initialise compartment volumes as a parameter - brett 20050908
        compartmentList = []
        for comp in m.getListOfCompartments():
            # print comp
            compartmentList.append((getName(comp), comp.getVolume()))

        if len(compartmentList) > 1:
            print('\nINFO: PySCeS models are assumed to have a single compartment')

        if len(compartmentList) > 0:
            F = open(self.model_filename + '.psc', 'a')
            F.write('\n## Initialise compartment volumes')
            for comp in compartmentList:
                F.write('\n' + comp[0] + ' = ' + str(comp[1]))
                ##  parameters.append(x[0])
                ##  init_par.append(x)
            F.write('\n')
            F.close()

        # Add assignment rules as forcing functions - brett 20050908
        pscRules = []
        for rule in m.getListOfRules():
            pscRules.append((rule.getVariable(), rule.getFormula()))

        if len(pscRules) > 0:
            F = open(self.model_filename + '.psc', 'a')
            F.write('\n## Assignment rules translated to forcing functions\n')
            for rule in pscRules:
                rule0 = 'self.' + rule[0]
                rule1l = rule[1].split()
                for word in range(len(rule1l)):
                    if rule1l[word].isalnum():
                        if rule1l[word] not in [
                            '1',
                            '2',
                            '3',
                            '4',
                            '5',
                            '6',
                            '7',
                            '8',
                            '9',
                        ]:
                            rule1l[word] = 'self.' + rule1l[word]
                F.write('!F ' + rule0 + ' = ')
                for word in rule1l:
                    F.write(word + ' ')
                F.write('\n')
            F.write('\n')
            F.close()

        if len(m.getNotes()) > 0:
            F = open(self.model_filename + '.psc', 'a')
            F.write('\n## Model notes' + m.getNotes().replace('\n', '\n# ') + '\n\n')
            F.close()

    def PSC_writeHeader(self, File):
        """
        PSC_writeHeader(File)

        Write a PSC file header to an open file object

        Arguments:
        =========
        File: a writable open text file object

        """
        try:
            UseR = getuser()
        except:
            UseR = ''

        header = ''
        # header +=  '############################################################\n'
        header += '# PySCeS (' + __version__ + ') model input file\n'
        header += '# PySCeS can be found at http://pysces.sourceforge.net/\n'
        # header += '###########################################################\n\n'
        header += '# Original input file: ' + File.name.split('\\')[-1][:-4] + '\n'
        header += (
            '# This file generated: '
            + strftime("%a, %d %b %Y %H:%M:%S")
            + ' by '
            + UseR
            + '\n\n'
        )
        File.write(header)
        File.write('\n')

    def PSC_writeSpecies(self, File, species):
        """
        PSC_writeSpecies(File,species)

        Write out model species initiaiisations to file

        Arguments:
        =========
        File: a writable open file object
        species: a list of (species.value) pairs

        """
        out_list = []

        out_list.append('\n## Variable species initial values\n')
        for x in range(len(species)):
            out_list.append(
                species[x][0] + ' = ' + self.mode_number_format % species[x][1] + '\n'
            )
        for x in out_list:
            File.write(x)
        File.write('\n')

    def PSC_writeFixedSpeciesList(self, File, fixed_species):
        """
        PSC_writeFixedSpeciesList(File,fixed_species)

        Write fixed species declaration to a PSC file

        Arguments:
        =========
        File: open, writable file object
        fixed_species: a list of (species,value) pairs

        """
        File.write('## Fixed species\n')
        if len(fixed_species) == 0:
            File.write('#  <none>')
        else:
            File.write('FIX: ')
            for x in fixed_species:
                File.write(x[0] + ' ')
        File.write('\n\n')

    def PSC_writeFixedSpecies(self, File, fixed_species):
        """
        PSC_writeFixedSpecies(File,fixed_species)

        Write fixed species initialisations to a PSC file

        Arguments:
        =========
        File: open, writable file object
        fixed_species: a list of (species,value) pairs

        """
        out_list = []

        out_list.append('\n## Fixed species\n')
        for x in range(len(fixed_species)):
            out_list.append(
                fixed_species[x][0]
                + ' = '
                + self.mode_number_format % fixed_species[x][1]
                + '\n'
            )
        for x in out_list:
            File.write(x)
        File.write('\n')

    def PSC_writeParameters(self, File, parameters):
        """
        PSC_writeParameters(File,parameters)

        Write mode parameter initialisations to a PSC file

        Arguments:
        =========
        File: open, writable file object
        parameters: a list of (parameter,value) pairs

        """
        out_list = []

        out_list.append('\n## Parameters\n')
        for x in range(len(parameters)):
            out_list.append(
                parameters[x][0]
                + ' = '
                + self.mode_number_format % parameters[x][1]
                + '\n'
            )
        for x in out_list:
            File.write(x)
        File.write('\n')

    def PSC_writeRateEquations(self, File, NetworkDict, number_format='%2.3f'):
        """
        PSC_writeRateEquations(File,NetworkDict,number_format='%2.3f')

        Write model rate equations to a PSC file

        Arguments:
        =========
        File: open, writable file object
        NetworkDict: a PySCeS network dictionary
        number_format [default='%2.3f']: number formatting to use in rate laws

        """
        out_list = []

        out_list.append('\n## Reaction stoichiometry and rate equations\n')
        for key in NetworkDict:
            out_list.append(key + ':\n')
            reagL = []
            reagR = []
            for reagent in NetworkDict[key]['Reagents']:
                if NetworkDict[key]['Reagents'][reagent] > 0:
                    if NetworkDict[key]['Reagents'][reagent] == 1.0:
                        reagR.append(reagent.replace('self.', ''))
                    else:
                        reagR.append(
                            '{'
                            + number_format % abs(NetworkDict[key]['Reagents'][reagent])
                            + '}'
                            + reagent.replace('self.', '')
                        )
                elif NetworkDict[key]['Reagents'][reagent] < 0:
                    if NetworkDict[key]['Reagents'][reagent] == -1.0:
                        reagL.append(reagent.replace('self.', ''))
                    else:
                        reagL.append(
                            '{'
                            + number_format % abs(NetworkDict[key]['Reagents'][reagent])
                            + '}'
                            + reagent.replace('self.', '')
                        )
                elif NetworkDict[key]['Reagents'][reagent] == 0:
                    # reagL.append(reagent.replace('self.',''))
                    print(NetworkDict[key]['Reagents'])
                    input('WTF: please contact developers')

            if len(reagL) == 0:
                print('Zero pool substrate', File.name)
                reagL.append('$pool')
            if len(reagR) == 0:
                print('Zero pool product', File.name)
                reagR.append('$pool')

            substring = ''
            count = 0
            for x in reagL:
                if count != 0:
                    substring += ' + '
                substring += x.replace(' ', '')
                count += 1
            prodstring = ''
            count = 0
            for x in reagR:
                if count != 0:
                    prodstring += ' + '
                prodstring += x.replace(' ', '')
                count += 1

            if NetworkDict[key]['Type'] == 'Rever':
                symbol = ' = '
            else:
                symbol = ' > '
            out_list.append('\t' + substring + symbol + prodstring + '\n')
            out_list.append(
                '\t' + NetworkDict[key]['RateEq'].replace('self.', '') + '\n\n'
            )
        for x in out_list:
            File.write(x)



"""
PySCeS - Python Simulator for Cellular Systems (http://pysces.sourceforge.net)

Copyright (C) 2004-2022 B.G. Olivier, J.M. Rohwer, J.-H.S Hofmeyr all rights reserved,

Brett G. Olivier (bgoli@users.sourceforge.net)
Triple-J Group for Molecular Cell Physiology
Stellenbosch University, South Africa.

Permission to use, modify, and distribute this software is given under the
terms of the PySceS (BSD style) license. See LICENSE.txt that came with
this distribution for specifics.

NO WARRANTY IS EXPRESSED OR IMPLIED.  USE AT YOUR OWN RISK.
Brett G. Olivier
"""
from __future__ import division, print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from pysces.version import __version__

__doc__ = '''pysx contributed modules interface'''

# load any unofficial user contrib stuff
from .PyscesContribUser import *

# load official contrib stuff
import os, sys, fileinput
from . import contrib as CONTRIB_module
from . import PyscesWeb

try:
    from . import PyscesSBML

    sbmlOK = 1
    libsbml = PyscesSBML.libsbml
except Exception as e:
    print('contrib: sbml not available', e)
    sbmlOK = 0


class CONTRIB_demo:
    def run(self):
        demo = CONTRIB_module.demo.test()
        demo.run()


class CONTRIB_visualise(PyscesSBML.PyscesSBML, PyscesWeb.PyscesHTML):
    """Visualise PySCeS models with PySBML"""

    def __init__(self):
        assert sbmlOK, "visualisation requires access to (lib)sbml"
        if os.sys.platform == 'win32':
            self.image_format = 'png'
        else:
            self.image_format = 'jpg'

    def static(self, mod, filename, slvl=2, dir=None):
        """
        static(mod,filename,slvl=2,dir=None)

        Generate a png/jpg image of a PySCeS model object

        Arguments:
        =========
        mod: a loaded PySCeS model
        filename: image file name
        slvl [default=2]: sbml level to use
        dir [default=None]: output directory

        """
        assert self.image_format in [
            'png',
            'jpg',
            'gif',
        ], 'Currently supported formats are:\n jpg, png, gif'
        if dir != None:
            filename = os.path.join(dir, filename)
        self.SBML_createModel(mod, filename, slvl, dir)
        self.SBML_setCompartment(mod.ModelFile[:-4].lower())
        self.SBML_setSpecies()
        self.SBML_setReactions()
        self.SBML_setModel()
        visMod = CONTRIB_module.visualise.VisualiseSBML(self.sbml_document.getModel())
        jpg = visMod.visualiseModel().static(self.image_format)
        F = open(filename + '.' + self.image_format, 'wb')
        F.write(jpg)
        F.close()

    def static_image_return(self, mod, slvl=2):
        """
        returns a binary image
        """
        assert self.image_format in [
            'png',
            'jpg',
            'gif',
        ], 'Currently supported formats are:\n jpg, png, gif'
        filename = 'dbasetemp'
        dir = None
        self.SBML_createModel(mod, filename, slvl, dir)
        self.SBML_setCompartment(mod.ModelFile[:-4].lower())
        self.SBML_setSpecies()
        self.SBML_setReactions()
        self.SBML_setModel()
        visMod = CONTRIB_module.visualise.VisualiseSBML(self.sbml_document.getModel())
        jpg = visMod.visualiseModel().static(self.image_format)

        return jpg

    def web(self, mod, filename, slvl=2, dir=None):
        """
        web(mod,filename,slvl=2,dir=None)

        Generate a png/jpg image of a PySCeS model object embedded in an HTML file

        Arguments:
        =========
        mod: a loaded PySCeS model
        filename: image/html file name
        slvl [default=2]: sbml level to use
        dir [default=None]: output directory

        """
        self.SBML_createModel(mod, filename, slvl, dir)
        if dir != None:
            filename = os.path.join(dir, filename)
        self.SBML_setCompartment(mod.ModelFile[:-4].lower())
        self.SBML_setSpecies()
        self.SBML_setReactions()
        self.SBML_setModel()
        visMod = CONTRIB_module.visualise.VisualiseSBML(self.sbml_document.getModel())
        jpg = visMod.visualiseModel().web(self.image_format)
        F = open(filename + '.' + self.image_format, 'wb')
        F.write(jpg[0])
        F.close()

        F2 = open(filename + '.html', 'w')
        F2 = self.HTML_header(F2)

        F2.write(
            '<img src="'
            + filename
            + '.'
            + self.image_format
            + '" usemap="#'
            + filename
            + '" border="0" alt="PySces image map">\n'
        )
        F2.write('<map name="' + filename + '">\n')
        F2.write(jpg[1])
        F2.write('</map>\n')
        F2 = self.HTML_footer(F2)
        F2.close()


class CONTRIB_sbw:
    """PySCeS interface to the Systems Biology Workbench"""

    def __init__(self):
        assert sbmlOK, "SBW requires access to (lib)sbml"
        self.StructAnalysis = CONTRIB_module.sbw.StructAnalysis()
        # self.MultipleViewsOfTimeSeriesData = CONTRIB_module.sbw.MultipleViewsOfTimeSeriesData()

    def LoadDrawNetwork(self):
        self.DrawNetworkGUI = CONTRIB_module.sbw.DrawNetworkGUI()


class contrib:
    """Dynamic on-availability module loading framework"""

    def __init__(self):
        for module in CONTRIB_module.mod_dict:
            if CONTRIB_module.mod_dict[module]['status']:
                loadOK = 2
                # print mod
                print('Loading module \"' + module + '\"')
                # exec('code = '+ CONTRIB_module.mod_dict[mod]['base']+'()')
                try:
                    setattr(
                        self,
                        module,
                        eval(CONTRIB_module.mod_dict[module]['base'] + '()'),
                    )
                    loadOK -= 1
                except Exception as f:
                    print(' Unable to instantiate module:', f)
                try:
                    setattr(self, '_' + module, eval('CONTRIB_module.' + module))
                    loadOK -= 1
                except Exception as f:
                    print(' Unable to load module:', f)
                if not loadOK:
                    print('ok.')
                else:
                    print(' Module \"' + module + '\" loading incomplete')
        print(' ')


del PyscesSBML, PyscesWeb



"""
PySCeS - Python Simulator for Cellular Systems (http://pysces.sourceforge.net)

Copyright (C) 2004-2022 B.G. Olivier, J.M. Rohwer, J.-H.S Hofmeyr all rights reserved,

Brett G. Olivier (bgoli@users.sourceforge.net)
Triple-J Group for Molecular Cell Physiology
Stellenbosch University, South Africa.

Permission to use, modify, and distribute this software is given under the
terms of the PySceS (BSD style) license. See LICENSE.txt that came with
this distribution for specifics.

NO WARRANTY IS EXPRESSED OR IMPLIED.  USE AT YOUR OWN RISK.
Brett G. Olivier
"""

from __future__ import division, print_function
from __future__ import absolute_import
from __future__ import unicode_literals


__version__ = '0.4.0'
__doc__ = 'PySCeS contrib module loader'


import os
import pysces

tempdir = os.getcwd()
mod_path = os.path.join(pysces.install_dir, 'contrib')

os.chdir(mod_path)
mod_dir_list = [dir for dir in os.listdir(mod_path) if os.path.isdir(dir)]
mod_load_list = []
mod_dict = {}

mod_os = os.name
# mod_os = 'posix' # transform a windows machine into a posix one

print('\nAdding contrib modules ...')
for mod in mod_dir_list:
    status = 0
    author = ''
    email = ''
    base = ''
    web = ''
    try:
        exec('import ' + mod + ' as CurMod')
        author = getattr(CurMod, 'pysx_author')
        email = getattr(CurMod, 'pysx_email')
        base = getattr(CurMod, 'pysx_base_class')
        web = getattr(CurMod, 'pysx_web')
        if mod_os in getattr(CurMod, 'pysx_oscompat'):
            print('Including module \"' + mod + '\"')
            if email != '':
                print('\tAuthor(s): ' + author + ', (' + email + ')')
            else:
                print('\tAuthor(s): ' + author)
            if web != '':
                print('\t' + web)
            status = 1
            mod_load_list.append(mod)
        else:
            print(
                '\t'
                + getattr(CurMod, 'pysx_name')
                + ' only available on '
                + str(getattr(CurMod, 'pysx_oscompat'))
            )
            print('\tMaintainer ' + author + email)
            status = 0
        del CurMod
    except Exception as e:
        print('\nModule ' + mod + ' *not* included')
        print('\t', e, '\n')
    mod_dict.update({mod: {}})
    mod_dict[mod].update(
        {
            'status': status,
            'path': os.path.abspath(os.path.join(mod_path, mod, '__init__.py')),
            'author': author,
            'email': email,
            'base': base,
        }
    )
print(' ')

os.chdir(tempdir)

del pysces, os, tempdir



# Automatic first-order derivatives
#
# Written by Konrad Hinsen <hinsen@cnrs-orleans.fr>
# last revision: 2002-6-13
#

# Adapted slightly to work independently and use numpy - bgoli
# Brett Olivier 20070308

"""This module provides automatic differentiation for functions with
any number of variables. Instances of the class DerivVar represent the
values of a function and its partial derivatives with respect to a
list of variables. All common mathematical operations and functions
are available for these numbers.  There is no restriction on the type
of the numbers fed into the code; it works for real and complex
numbers as well as for any Python type that implements the necessary
operations.

This module is as far as possible compatible with the n-th order
derivatives module Derivatives. If only first-order derivatives
are required, this module is faster than the general one.

Example:

  >>>print sin(DerivVar(2))

  produces the output

  >>>(0.909297426826, [-0.416146836547])

The first number is the value of sin(2); the number in the following
list is the value of the derivative of sin(x) at x=2, i.e. cos(2).

When there is more than one variable, DerivVar must be called with
an integer second argument that specifies the number of the variable.

Example:

  >>>x = DerivVar(7., 0)
  >>>y = DerivVar(42., 1)
  >>>z = DerivVar(pi, 2)
  >>>print (sqrt(pow(x,2)+pow(y,2)+pow(z,2)))

  produces the output

  >>>(42.6950770511, [0.163953328662, 0.98371997197, 0.0735820818365])

The numbers in the list are the partial derivatives with respect
to x, y, and z, respectively.

Note: It doesn't make sense to use DerivVar with different values
for the same variable index in one calculation, but there is
no check for this. I.e.

  >>>print DerivVar(3, 0)+DerivVar(5, 0)

  produces

  >>>(8, [2])

but this result is meaningless.
"""
from __future__ import division, print_function
from __future__ import absolute_import
from __future__ import unicode_literals

# Changed Numeric to Numpy and updated code to reflect this - bgoli

# import Numeric
try:
    import numpy
except Exception as ex:
    print(ex)
    print('numpy import failed trying to import numeric (this is not ideal) ... ')
    import Numeric as numpy

# The following class represents variables with derivatives:

class DerivVar:

    """Variable with derivatives

    Constructor: DerivVar(|value|, |index| = 0)

    Arguments:

    |value| -- the numerical value of the variable

    |index| -- the variable index (an integer), which serves to
               distinguish between variables and as an index for
               the derivative lists. Each explicitly created
               instance of DerivVar must have a unique index.

    Indexing with an integer yields the derivatives of the corresponding
    order.
    """

    def __init__(self, value, index=0, order=1):
        if order > 1:
            raise ValueError('Only first-order derivatives')
        self.value = value
        if order == 0:
            self.deriv = []
        elif type(index) == type([]):
            self.deriv = index
        else:
            self.deriv = index*[0] + [1]

    def __getitem__(self, item):
        if item < 0 or item > 1:
            raise ValueError('Index out of range')
        if item == 0:
            return self.value
        else:
            return self.deriv

    def __repr__(self):
        return repr((self.value, self.deriv))

    def __str__(self):
        return str((self.value, self.deriv))

    def __coerce__(self, other):
        if isDerivVar(other):
            return self, other
        else:
            return self, DerivVar(other, [])

    def __cmp__(self, other):
        return cmp(self.value, other.value)

    def __neg__(self):
        return DerivVar(-self.value,[-a for a in self.deriv])

    def __pos__(self):
        return self

    def __abs__(self): # cf maple signum # derivate of abs
        absvalue = abs(self.value)
        return DerivVar(absvalue, list(map(lambda a, d=self.value/absvalue:
                                      d*a, self.deriv)))
    def __bool__(self):
        return self.value != 0

    def __add__(self, other):
        return DerivVar(self.value + other.value,
                _mapderiv(lambda a,b: a+b, self.deriv, other.deriv))

    __radd__ = __add__

    def __sub__(self, other):
        return DerivVar(self.value - other.value,
            _mapderiv(lambda a,b: a-b, self.deriv, other.deriv))

    def __rsub__(self, other):
        return DerivVar(other.value - self.value,
            _mapderiv(lambda a,b: a-b, other.deriv, self.deriv))

    def __mul__(self, other):
        return DerivVar(self.value*other.value,
            _mapderiv(lambda a,b: a+b,
                  list(map(lambda x,f=other.value:f*x, self.deriv)),
                  list(map(lambda x,f=self.value:f*x, other.deriv))))

    __rmul__ = __mul__

    def __div__(self, other):
        if not other.value:
            raise ZeroDivisionError('DerivVar division')
        inv = 1./other.value
        return DerivVar(self.value*inv,
                _mapderiv(lambda a,b: a-b,
                      list(map(lambda x,f=inv: f*x, self.deriv)),
                      list(map(lambda x,f=self.value*inv*inv: f*x,
                          other.deriv))))

    def __rdiv__(self, other):
        return other/self

    def __pow__(self, other, z=None):
        if z is not None:
            raise TypeError('DerivVar does not support ternary pow()')
        val1 = pow(self.value, other.value-1)
        val = val1*self.value
        deriv1 = list(map(lambda x,f=val1*other.value: f*x, self.deriv))
        if isDerivVar(other) and len(other.deriv) > 0:
            deriv2 = list(map(lambda x, f=val*numpy.log(self.value): f*x,
                             other.deriv))
            return DerivVar(val,_mapderiv(lambda a,b: a+b, deriv1, deriv2))
        else:
            return DerivVar(val,deriv1)

    def __rpow__(self, other):
        return pow(other, self)

    def exp(self):
        v = numpy.exp(self.value)
        return DerivVar(v, list(map(lambda x,f=v: f*x, self.deriv)))

    def log(self):
        v = numpy.log(self.value)
        d = 1./self.value
        return DerivVar(v, list(map(lambda x,f=d: f*x, self.deriv)))

    def log10(self):
        v = numpy.log10(self.value)
        d = 1./(self.value * numpy.log(10))
        return DerivVar(v, list(map(lambda x,f=d: f*x, self.deriv)))

    def sqrt(self):
        v = numpy.sqrt(self.value)
        d = 0.5/v
        return DerivVar(v, list(map(lambda x,f=d: f*x, self.deriv)))

    def sign(self):
        if self.value == 0:
            raise ValueError("can't differentiate sign() at zero")
        return DerivVar(numpy.sign(self.value), 0)

    def sin(self):
        v = numpy.sin(self.value)
        d = numpy.cos(self.value)
        return DerivVar(v, list(map(lambda x,f=d: f*x, self.deriv)))

    def cos(self):
        v = numpy.cos(self.value)
        d = -numpy.sin(self.value)
        return DerivVar(v, list(map(lambda x,f=d: f*x, self.deriv)))

    def tan(self):
        v = numpy.tan(self.value)
        d = 1.+pow(v,2)
        return DerivVar(v, list(map(lambda x,f=d: f*x, self.deriv)))

    def sinh(self):
        v = numpy.sinh(self.value)
        d = numpy.cosh(self.value)
        return DerivVar(v, list(map(lambda x,f=d: f*x, self.deriv)))

    def cosh(self):
        v = numpy.cosh(self.value)
        d = numpy.sinh(self.value)
        return DerivVar(v, list(map(lambda x,f=d: f*x, self.deriv)))

    def tanh(self):
        v = numpy.tanh(self.value)
        d = 1./pow(numpy.cosh(self.value),2)
        return DerivVar(v, list(map(lambda x,f=d: f*x, self.deriv)))

    def arcsin(self):
        v = numpy.arcsin(self.value)
        d = 1./numpy.sqrt(1.-pow(self.value,2))
        return DerivVar(v, list(map(lambda x,f=d: f*x, self.deriv)))

    def arccos(self):
        v = numpy.arccos(self.value)
        d = -1./numpy.sqrt(1.-pow(self.value,2))
        return DerivVar(v, list(map(lambda x,f=d: f*x, self.deriv)))

    def arctan(self):
        v = numpy.arctan(self.value)
        d = 1./(1.+pow(self.value,2))
        return DerivVar(v, list(map(lambda x,f=d: f*x, self.deriv)))

    def arctan2(self, other):
        den = self.value*self.value+other.value*other.value
        s = self.value/den
        o = other.value/den
        return DerivVar(numpy.arctan2(self.value, other.value),
                _mapderiv(lambda a,b: a-b,
                      list(map(lambda x,f=o: f*x, self.deriv)),
                      list(map(lambda x,f=s: f*x, other.deriv))))

# Don't know if this will work - bgoli
    ##  def gamma(self):
        ##  from transcendental import gamma, psi
        ##  v = gamma(self.value)
        ##  d = v*psi(self.value)
        ##  return DerivVar(v, map(lambda x,f=d: f*x, self.deriv))

# Type check

def isDerivVar(x):
    "Returns 1 if |x| is a DerivVar object."
    return hasattr(x,'value') and hasattr(x,'deriv')

# Map a binary function on two first derivative lists

def _mapderiv(func, a, b):
    nvars = max(len(a), len(b))
    a = a + (nvars-len(a))*[0]
    b = b + (nvars-len(b))*[0]
    return list(map(func, a, b))


# Don't use this - bgoli
##  def DerivVector(x, y, z, index=0):

    ##  """Returns a vector whose components are DerivVar objects.

    ##  Arguments:

    ##  |x|, |y|, |z| -- vector components (numbers)

    ##  |index| -- the DerivVar index for the x component. The y and z
               ##  components receive consecutive indices.
    ##  """

    ##  from Scientific.Geometry.Vector import Vector
    ##  if isDerivVar(x) and isDerivVar(y) and isDerivVar(z):
        ##  return Vector(x, y, z)
    ##  else:
        ##  return Vector(DerivVar(x, index),
                      ##  DerivVar(y, index+1),
                      ##  DerivVar(z, index+2))



"""
PySCeS - Python Simulator for Cellular Systems (http://pysces.sourceforge.net)

Copyright (C) 2004-2022 B.G. Olivier, J.M. Rohwer, J.-H.S Hofmeyr all rights reserved,

Brett G. Olivier (bgoli@users.sourceforge.net)
Triple-J Group for Molecular Cell Physiology
Stellenbosch University, South Africa.

Permission to use, modify, and distribute this software is given under the
terms of the PySceS (BSD style) license. See LICENSE.txt that came with
this distribution for specifics.

NO WARRANTY IS EXPRESSED OR IMPLIED.  USE AT YOUR OWN RISK.
Brett G. Olivier
"""

from __future__ import division, print_function
from __future__ import absolute_import
from __future__ import unicode_literals

'''
This file is for "unofficial" user contributed modules. It is imported into PyscesContrib using
 from PyscesContribUser import *
Make sure to use the same module class definitions as used in PyscesContrib i.e. CONTRIB_modulename
to allow for automated module loading and instantiation.
'''

from pysces.version import __version__

__doc__ = '''pysx unofficial user contributed module interface'''



import os
import itertools

os.chdir('c:/mypysces/kraken')
##  os.chdir('/home/bgoli/mypysces/kraken')
##  import pysces
from . import *

##  from Kraken import *


# investigate using vstack with type/range checking
def concatenateArrays(array_list):
    output = None
    for arr in range(len(array_list)):
        if type(array_list[arr]) == numpy.ndarray:
            if arr == 0:
                output = array_list[arr]
            else:
                output = scipy.concatenate((output, array_list[arr]))
    return output


def buildCycler(lst):
    """
    creates an infinite looping generator with itertools
    """
    return itertools.cycle(lst)


def deltaCommand():
    """
    This is used to setup conditions for each next-job-on-server
    """
    delta = 'P_NONE'
    return [delta]


def Scheduler1(contrl, job_list, init_list, task_id):
    """
    contrl is a getBasicController() instance
    job_list is a list of jobs to run
    init_list set of `static' command used to initialise each job
    """
    original_job_list = tuple(job_list)
    TIME_START = time.time()
    DaResults = []
    JobsWaiting = True
    getServer = buildCycler(contrl.available_server_list)
    while JobsWaiting:
        deferred_jobs = []
        job_servers = []
        for job in range(len(job_list)):
            if job < len(contrl.available_server_list):
                print('\nJob %s queued for processing...' % (job + 1))
                server = next(getServer)
                job_servers.append(server)
                print("\nProcessing job %s init list ..." % (job + 1))
                contrl.sendCmdListToOne(init_list, server)
                print("\nProcessing job %s delta command ..." % (job + 1))
                contrl.sendCmdListToOne(deltaCommand(), server)
                contrl.sendJobToOne(job_list[job], server)
            else:
                deferred_jobs.append(job_list[job])
                print('Job %s deferred and rescheduled.' % (job + 1))
        print("\nProcessing queued jobs ...")
        contrl.runAllJobs()
        for server in job_servers:
            contrl.getDataFromOneJob(server)
            arr = contrl.tentacle_response[server]
            DaResults.append(arr)
            print(type(arr))
        contrl.clearProcessMap()
        if len(deferred_jobs) > 0:
            print('\nDeferred:', deferred_jobs)
            job_list = deferred_jobs
        else:
            JobsWaiting = False
        cPickle.dump(DaResults, open(task_id + '_result_list.bin', 'wb'), 2)

    TIME_END = time.time()

    print('\nNumber of results = %s' % len(DaResults))
    total_states = 0
    for x in range(len(DaResults)):
        try:
            print(
                '\tResult %s has %s rows and %s columns'
                % (x + 1, DaResults[x][0].shape[0], DaResults[x][0].shape[1])
            )
            total_states += DaResults[x][0].shape[0]
        except:
            print(type(DaResults[x][0]))
            print(
                '\tResult %s %s: %s' % (x + 1, type(DaResults[x][0]), DaResults[x][0])
            )
            print(original_job_list[x])

    print(
        'Total time taken to complete %s state scan %s minutes'
        % (total_states, (TIME_END - TIME_START) / 60.0)
    )
    return DaResults


def GridSortLR(arr):
    """
    Sort irregular unstructured grid data into monotonically rising grid data
    (where possible) Work from left column to right (uses ultra cool NumPy functions :-)
    """
    print('arr.shape:', arr.shape)
    sorted = numpy.rot90(arr)
    print('rotated.shape:', sorted.shape)
    sorted_idx = numpy.lexsort(sorted)
    print('sorted_idx.shape', sorted_idx.shape)
    sorted = sorted.take(sorted_idx, axis=-1)
    sorted = numpy.flipud(sorted)
    arr = numpy.transpose(sorted)
    print('arr.shape:', arr.shape)
    return arr


def ScalarFunc(val):
    """
    Manipulate the scalar (or z-axis) values
    Accepts and returns a double
    """
    return val


def writeVTK_UnstructuredGrid(fname, arr):
    """
    Write arr as a VTK unstructured grid to file "fname.vtk"
    arr must either be in:
        x,y,z format (in which case z is used for scalar data) or
        z,y,z,s format (s being point scalar value)
    Overwrite the ScalarFunc(val) function to manipulate the scalar (z-axis) values

    Original version http://mayavi.sourceforge.net/mwiki/PointClouds
    Hacked quite a bit by brett 20070221
    """
    assert arr.shape[1] == 3 or arr.shape[1] == 4, '\nneed s or for columns for this'
    if arr.shape[1] == 4:
        HAVE_SCALARS = 1
    else:
        HAVE_SCALARS = 0
        print('No scalar values supplied Z axis values will be used')

    n = arr.shape[0]
    print("n:", n)
    # write data to vtk polydata file
    # write header
    out = open(fname + '.vtk', 'w')
    h1 = "# vtk DataFile Version 2.0\n"
    h1 += "%s\n" % fname
    h1 += "ASCII\n"
    h1 += "DATASET UNSTRUCTURED_GRID\n"
    h1 += "POINTS " + str(n) + " double\n"
    out.write(h1)
    # write xyz data
    for r in range(n):
        # s = '%15.2f %15.2f %15.2f' % (x[i], y[i], z[i])
        out.write(str(arr[r, 0]) + " " + str(arr[r, 1]) + " " + str(arr[r, 2]) + '\n')

    # write cell data
    out.write("CELLS " + str(n) + " " + str(2 * n) + '\n')
    for r in range(n):
        # s = '1 %d \n' % (i)
        out.write("1 " + str(r) + "\n")

    # write cell types
    out.write("CELL_TYPES " + str(n) + '\n')
    for r in range(n):
        out.write("1 \n")

    # write z scalar values
    h2 = '\n' + """POINT_DATA """ + str(n) + "\n"
    h3 = "SCALARS %s double 1\n" % fname
    h3 += "LOOKUP_TABLE default\n"
    out.write(h2 + h3)

    for r in range(n):
        if HAVE_SCALARS:
            sc = ScalarFunc(arr[r, 3])
        else:
            sc = ScalarFunc(arr[r, 2])
        out.write(str(sc) + "\n")

    out.write('\n')
    out.close()


if __name__ == '__main__':
    import scipy.io

    """
    Build controller and initialise active tentacles
    """
    blob = getBasicController()
    Model_File = '2006_isola2.psc'
    Model_Dir = 'c:\\mypysces\\pscmodels\\isola06'
    ##  Model_Dir = '/home/bgoli/mypysces/pscmodels/isola06'
    blob.startModelServer(Model_File, Model_Dir)
    blob.startTentacleMonitor(interval=60)
    print('Server list: %s\n' % blob.available_server_list)

    """
    Get a name for this task (task == collection of jobs)
    """
    task_name = 'cont_eigen'

    """
    Set up a list of 'static' initialisation commands that will be used to
    initialise each server before a job is sent to it
    """
    I_list = [
        'P_INIT,' + blob.model_server.model_file_name,
        'P_LOAD',
        'P_SET_FLOAT,mode_state_init,3',
        'P_SET_FLOAT,nleq2_nitmax,20',
        'P_SET_FLOAT,mode_state_init3_factor,0.5',
        'P_SET_FLOAT,pitcon_iter,20',
        'P_SET_FLOAT,V2,200.0',
        'P_SET_FLOAT,V3,200.0',
        'P_SET_FLOAT,S05,1.0',
    ]
    ##  'P_SET_FLOAT,V2,1000.0', 'P_SET_FLOAT,V3,1000.0', 'P_SET_FLOAT,S05,1.0']

    """
    Set up our continuation:
    We want a parameter to be set with every new job/server and make use
    of the deltaCommand() function. Here we use it to set our y-axis (A05) values
    """
    yrange = 100
    xpoints = 256

    # create a list of values
    ##  delta_list_vl = scipy.logspace(scipy.log10(0.1),scipy.log10(5.0), yrange)
    ##  delta_list_l = scipy.logspace(scipy.log10(5.1),scipy.log10(100.0), yrange)
    ##  delta_list_h = scipy.logspace(scipy.log10(100.1),scipy.log10(5000.0), yrange*3)
    ##  delta_list = scipy.hstack((delta_list_vl, delta_list_l, delta_list_h))
    delta_list = scipy.logspace(scipy.log10(0.1), scipy.log10(1.0), yrange)
    # initiate an 'looping' generator with it
    delta_gen = buildCycler(delta_list)

    def deltaCommand():
        """
        This is used to setup conditions for each next-job-on-server
        Used in sendCmdListToOne() so may one or more commands returned
        as a list
        """
        delta = 'P_SET_FLOAT,A05,%s' % next(delta_gen)
        # must return a list !!!
        return [delta]

    """
    Set up our list of jobs, in this case a single continuation+eigen with
    density xpoints at each of the delta (A05 values)
    """
    J_list = []
    for d in delta_list:
        ##  J_list.append('P_CONTINUATION,P,0.001,4.0e2,%s,%s' % (xpoints,d))
        J_list.append('P_CONTINUATION_WITH_EIGEN,P,0.01,3.0e2,%s,%s' % (xpoints, d))

    """
    Feed the controller, job list, initiation list and task name to the scheduler
    and off we go, we end up with a list of results
    """
    DaResults = Scheduler1(blob, J_list, I_list, task_name)

    """
    From here on we are processing/formatting our result data
    """
    iRes = []
    sRes = []
    jRes = []
    eRes = []
    for res in DaResults:
        iRes.append(res[0])
        sRes.append(res[1])
        jRes.append(res[2])
        eRes.append(res[3])
    del DaResults

    FinalIArray = concatenateArrays(iRes)
    print('Indexes[0]', FinalIArray[0])
    FinalSArray = concatenateArrays(sRes)
    FinalJArray = concatenateArrays(jRes)
    FinalEArray = concatenateArrays(eRes)
    print(FinalIArray.shape, FinalIArray.dtype)
    print(FinalSArray.shape, FinalSArray.dtype)
    print(FinalJArray.shape, FinalJArray.dtype)
    print(FinalEArray.shape, FinalEArray.dtype)
    del iRes
    del sRes
    del jRes
    del eRes

    cPickle.dump(
        (FinalIArray, FinalSArray, FinalJArray, FinalEArray),
        open(task_name + '_result_array.bin', 'wb'),
        2,
    )

    # get rid of negative values in axes and fluxes
    FinalArrayClean = []
    FinalVTK = []
    for r in range(FinalSArray.shape[0]):
        if (
            FinalIArray[r, :].all() > 0.0
            and FinalSArray[r, :].all() > 0.0
            and FinalJArray[r, :].all() > 0.0
            and FinalIArray[r, 1] >= 0.2
            and FinalJArray[r, 1] >= 0.08
            and FinalIArray[r, 1] <= 200.0
        ):
            FinalArrayClean.append(
                [
                    FinalIArray[r, 0],
                    FinalIArray[r, 1],
                    FinalJArray[r, 1],
                    max(FinalEArray[r, :].real),
                ]
            )
        else:
            FinalIArray[r, :] = 0.0
            FinalSArray[r, :] = 0.0
            FinalJArray[r, :] = 0.0
            FinalEArray[r, :] = 0.0
    FinalArrayClean = numpy.array(FinalArrayClean)

    # write out an xyz gplt file
    coords = numpy.transpose(
        numpy.array(
            (FinalArrayClean[:, 0], FinalArrayClean[:, 1], FinalArrayClean[:, 2])
        )
    )
    # get our data in nice grid format for plotting
    coords = GridSortLR(coords)
    scipy.io.write_array(task_name + '_result_gplt.dat', coords, separator=' ')

    def ScalarFunc(val):
        """
        Manipulate the scalar (z-axis) values
        Accepts and returns a double
        """
        if val >= 0.0:
            return 60.0
        else:
            return val

    # prepare vtk data
    vtkcoords = numpy.transpose(
        numpy.array(
            (
                FinalArrayClean[:, 0],
                FinalArrayClean[:, 1],
                FinalArrayClean[:, 2],
                FinalArrayClean[:, 3],
            ),
            'd',
        )
    )
    for row in range(vtkcoords.shape[0]):
        vtkcoords[row, :3] = scipy.log10(vtkcoords[row, :3])

    print(vtkcoords[:10, :])
    vtkcoords = GridSortLR(vtkcoords)
    print(vtkcoords[:10, :])

    writeVTK_UnstructuredGrid(task_name + '_result_ug', vtkcoords)



#!/usr/bin/env python

"""
PySCeS - Python Simulator for Cellular Systems (http://pysces.sourceforge.net)

Copyright (C) 2004-2022 B.G. Olivier, J.M. Rohwer, J.-H.S Hofmeyr all rights reserved,

Brett G. Olivier (bgoli@users.sourceforge.net)
Triple-J Group for Molecular Cell Physiology
Stellenbosch University, South Africa

Permission to use, modify, and distribute this software is given under the
terms of the PySceS (BSD style) license.  See LICENSE.txt that came with
this distribution for specifics.

NO WARRANTY IS EXPRESSED OR IMPLIED.  USE AT YOUR OWN RISK.
Brett G. Olivier
"""
from __future__ import division, print_function
from __future__ import absolute_import

__doc__ = "PySCeS: the Python Simulator for Cellular Systems setup file"

# get __version__ from version.txt
with open('pysces/version.txt') as f:
    __version__ = f.read().strip()

# avoid duplication
with open('requirements.txt') as f:
    requirements = f.read().splitlines()

import os, re

try:
    import configparser  # Py 3
except ImportError:
    import ConfigParser as configparser  # Py 2

try:
    print('Building an egg? %s.' % FRYING_EGGS)
except:
    FRYING_EGGS = False

try:
    import setuptools

    print('setuptools is available.')
except Exception as ex:
    print('setuptools not available.')

try:
    from numpy.distutils.core import setup, Extension
except Exception as ex:
    print(ex)
    print("PySCeS requires NumPy and SciPy 0.6x+\n")
    os.sys.exit(-1)

try:
    import distutils.command.bdist_conda

    print('bdist_conda is available.')
except Exception as ex:
    print('bdist_conda not available.')


########## User configuration section ##########

# Install extension modules (zero for skip module) - brett 20040310
pitcon = 1

##Special licence libraries see the relevant readme.txt file for details
nleq2 = 1  # pysces/nleq2/readme.txt

use = re.split('\s+', os.getenv('PYSCES_USE', ''))
for e in use:
    if e == 'pitcon':
        pitcon = 1
    elif e == 'nopitcon':
        pitcon = 0
    elif e == 'nleq2':
        nleq2 = 1
    elif e == 'nonleq2':
        nleq2 = 0
'''
# this is now obsolete with nleq2 4.3
    elif e.startswith('nleq2_byteorder='):
        nleq2_byteorder_override = int(e[16:])
# detects and uses IEEE fp big/little endian
'''
### End user configuration section

########## From here on it's up to distutils ##########

# get the dir of setup.py
local_path = os.path.dirname(os.path.abspath(os.sys.argv[0]))
os.chdir(local_path)

myscripts = []
mydata_files = []

# add some model files into pscmodels
modfold = os.path.join('pysces', 'pscmodels')
mods = os.listdir(modfold)
alist = []
for x in mods:
    if x[-4:] != '.psc':
        pass
    else:
        alist.append(os.path.join(modfold, x))
mydata_files.append((os.path.join('pysces', 'pscmodels'), alist))

# Default configurations for the pyscfg.ini files
if os.sys.platform == 'win32':
    if FRYING_EGGS:
        eggdir = 'pysces-%s-py%s.%s-%s.egg' % (
            __version__,
            os.sys.version_info[0],
            os.sys.version_info[1],
            os.sys.platform,
        )
        installdir = os.path.join(
            os.sys.prefix, 'lib', 'site-packages', eggdir, 'pysces'
        )
    else:
        installdir = os.path.join(os.sys.prefix, 'lib', 'site-packages', 'pysces')
    config = {
        "install_dir": installdir,
        "model_dir": os.path.join(os.getenv('USERPROFILE'), 'Pysces', 'psc'),
        "output_dir": os.path.join(os.getenv('USERPROFILE'), 'Pysces'),
        "gnuplot_dir": None,
        "silentstart": False,
        "change_dir_on_start": False,
    }
else:
    if hasattr(os.sys, 'lib'):
        lib = os.sys.lib
    else:
        lib = 'lib'
    config = {
        "install_dir": os.path.join(
            os.sys.prefix,
            lib,
            "python%d.%d" % tuple(os.sys.version_info[:2]),
            'site-packages',
            'pysces',
        ),
        "model_dir": os.path.join(os.path.expanduser('~'), 'Pysces', 'psc'),
        "output_dir": os.path.join(os.path.expanduser('~'), 'Pysces'),
        "gnuplot_dir": None,
        "silentstart": False,
        "change_dir_on_start": False,
    }


def writeConfig(local_path, config={}):
    cfgfile = open(os.path.join(local_path, 'pysces', 'pyscfg.ini'), 'w')
    cp = configparser.ConfigParser()
    # PySCeS internal setup
    cp.add_section('Pysces')
    for key in config:
        print(repr(key) + ' :: ' + str(config[key]))
        cp.set('Pysces', key, str(config[key]))
    # add configuration data
    cp.add_section('PyscesConfig')
    cp.set('PyscesConfig', 'matplotlib', 'True')
    # OSX patch thanks to AF
    if os.sys.platform == 'darwin':
        cp.set('PyscesConfig', 'matplotlib_backend', 'MacOSX')
    else:
        cp.set('PyscesConfig', 'matplotlib_backend', 'TkAgg')
    cp.set('PyscesConfig', 'gnuplot', 'False')
    # Built in modules
    cp.add_section('PyscesModules')
    if pitcon:
        cp.set('PyscesModules', 'pitcon', 'True')
    else:
        cp.set('PyscesModules', 'pitcon', 'False')
    # PySCeS external module setup
    cp.add_section('ExternalModules')
    if nleq2:
        cp.set('ExternalModules', 'nleq2', 'True')
        mydata_files.append(
            (
                os.path.join('pysces', 'nleq2'),
                [os.path.join('pysces', 'nleq2', 'nleq2_readme.txt')],
            )
        )
    else:
        cp.set('ExternalModules', 'nleq2', 'False')
        mydata_files.append(
            (
                os.path.join('pysces', 'nleq2'),
                [os.path.join('pysces', 'nleq2', 'readme.txt')],
            )
        )
    cp.write(cfgfile)
    cfgfile.close()


writeConfig(local_path, config)
print('Default configuration file installed')

# my subpackage list
mypackages = [
    'pysces',
    'pysces.tests',
    'pysces.lib',
    'pysces.pitcon',
    'pysces.sandbox',
    'pysces.contrib',
    'pysces.contrib.demo',
    'pysces.core2',
    'pysces.kraken',
    'pysces.kraken.controllers',
]

# PySCeS modules
mymodules = []

# extra compile args for FORTRAN code that should meand that gfortran libraries are statically compiled into the pyd
if os.sys.platform == 'win32':
    # extra_f77_compile_args = ["-static-libgcc", "-static-libgfortran"]
    extra_f77_compile_args = ["-static", "-static-libgcc", "-static-libgfortran"]
elif os.sys.platform.startswith('freebsd13'):
    # this is for future FreeBSD 13 compatability
    extra_f77_compile_args = ["-lpython3.8", "-shared"]
    #extra_link_args = ["-lpython3.8", "-shared"] # if the above doesn't work
else:
    extra_f77_compile_args = []

print('extra_f77_compile_args: ', extra_f77_compile_args)


if pitcon:
    print('\nBuilding pitcon')
    extpath = os.path.join('pysces', 'pitcon')
    pitcon = Extension(
        'pysces.pitcon.pitcon',
        sources=[
            os.path.join(extpath, 'pitcon.pyf'),
            os.path.join(extpath, 'pcon61subd.f'),
            os.path.join(extpath, 'dpcon61.f'),
            os.path.join(extpath, 'dpcon61w.f'),
        ],
        extra_f77_compile_args=extra_f77_compile_args,
    )
    mymodules.append(pitcon)
    # mydata_files.append((os.path.join('pysces','pitcon'), [os.path.join(local_path, 'pysces', 'pitcon','readme.txt'), os.path.join(local_path, 'pysces', 'pitcon','readme.txt')]))
else:
    print('\nSkipping pitcon')


if nleq2:
    print('\nBuilding nleq2')
    '''
    # this is now obsolete with nleq2 4.3 ... i hope !
    print 'System ByteOrder', os.sys.byteorder
    if os.path.exists(os.path.join(local_path, 'pysces', 'nleq2','nleq2.f')) and nleq2_byteorder_override:
        print 'INFO: using user supplied nleq2.f'
    else:
        if os.sys.byteorder == 'little':
            shutil.copyfile(os.path.join(extpath,'nleq2_little.f'), os.path.join(extpath,'nleq2.f'))
        elif os.sys.byteorder == 'big':
            shutil.copyfile(os.path.join(extpath,'nleq2_big.f'), os.path.join(extpath,'nleq2.f'))
    '''
    extpath = os.path.join('pysces', 'nleq2')
    nleq2 = Extension(
        'pysces.nleq2.nleq2',
        sources=[
            os.path.join(extpath, 'nleq2.pyf'),
            os.path.join(extpath, 'nleq2.f'),
            os.path.join(extpath, 'linalg_nleq2.f'),
            os.path.join(extpath, 'zibmon.f'),
            os.path.join(extpath, 'zibsec.f'),
            os.path.join(extpath, 'zibconst.f'),
            os.path.join(extpath, 'wnorm.f'),
        ],
        extra_f77_compile_args=extra_f77_compile_args,
    )
    mymodules.append(nleq2)
    mypackages.append('pysces.nleq2')
else:
    print('\n')

if len(mymodules) == 0:
    noext = Extension('None', [], None)  # Not ideal but seems safe
    mymodules.append(noext)

# Data files to copy
mydata_files.append(
    (
        os.path.join('pysces'),
        [os.path.join('pysces', 'pyscfg.ini'), os.path.join('pysces', 'version.txt')],
    )
)
mydata_files.append(('', [os.path.join('pysces', 'pysces.pth')]))

# JR 2021-08 userguide.pdf included as a symlink to built docs in submodule
mydata_files.append(
    (os.path.join("pysces", "docs"), [os.path.join("pysces", "docs", "userguide.pdf")])
)
mydata_files.append(
    (
        os.path.join('pysces', 'examples'),
        [
            os.path.join('pysces', 'examples', examplefile)
            for examplefile in os.listdir(
                os.path.join(local_path, 'pysces', 'examples')
            )
        ],
    )
)
## not sure if this is necessary anymore now that I am using static gfortran linking
if os.sys.platform == 'win32':
    mydata_files.append(
        (
            os.path.join('pysces', 'win32'),
            [
                os.path.join('pysces', 'win32', 'libquadmath-0.dll'),
                os.path.join('pysces', 'win32', 'libgfortran-3.dll'),
                os.path.join('pysces', 'win32', 'libgcc_s_seh-1.dll'),
                os.path.join('pysces', 'win32', 'libwinpthread-1.dll'),
            ],
        )
    )

os.chdir(local_path)
# Install packages and the metatool binaries as "data"
setup(
    name="pysces",
    version=__version__,
    description="The Python Simulator for Cellular Systems - simulation and analysis tools for modelling biological systems",
    long_description="""
 PySCeS is developed by the Triple-J Group for Molecular Cell Physiology
 in order to try model and understand the complex processes and systems
 which make up the living cell.

    PySCeS features, amongst other things:
    - A text based model description language.
    - A structural analysis module.
    - Integrators for time simulation
    - Non-linear solvers for steady-state analysis
    - A module for performing Metabolic Control Analysis
    - A bifurcation module for systems which exhibit multiple steady states
    - A variety of extra utilites for parameter scans, data output and plotting.
    - A dynamic module loading framework.
    - SBML import and export capability.
    """,
    author="Brett G. Olivier and Johann M. Rohwer",
    author_email="pysces@googlegroups.com",
    maintainer="Brett G. Olivier and Johann M. Rohwer",
    maintainer_email="pysces@googlegroups.com",
    url="http://pysces.sourceforge.net",
    download_url="https://pypi.org/project/pysces/#files",
    license="New BSD style",
    keywords="computational systems biology, modelling, simulation, systems biology",
    zip_safe=False,
    install_requires=requirements,
    extras_require={
        'parscan': ['ipyparallel'],
        'cvode': ['assimulo'],
        'sbml': ['python_libsbml'],
        'all': ['ipyparallel', 'assimulo', 'python_libsbml'],
    },
    platforms=["Windows", "POSIX", "Max OSX"],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Development Status :: 6 - Mature',
        'Environment :: Console',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Fortran',
        'Programming Language :: Python',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Scientific/Engineering :: Chemistry',
    ],
    packages=mypackages,
    data_files=mydata_files,
    ext_modules=mymodules,
)



"""
PySCeS - Python Simulator for Cellular Systems (http://pysces.sourceforge.net)

Copyright (C) 2004-2022 B.G. Olivier, J.M. Rohwer, J.-H.S Hofmeyr all rights reserved,

Brett G. Olivier (bgoli@users.sourceforge.net)
Triple-J Group for Molecular Cell Physiology
Stellenbosch University, South Africa.

Permission to use, modify, and distribute this software is given under the
terms of the PySceS (BSD style) license. See LICENSE.txt that came with
this distribution for specifics.

NO WARRANTY IS EXPRESSED OR IMPLIED.  USE AT YOUR OWN RISK.
Brett G. Olivier
"""
from __future__ import division, print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from .version import __version__
from . import __SILENT_START__
import subprocess, os, shutil, time, math, itertools, copy
from getpass import getuser
import numpy, scipy

try:
    input = raw_input  # Py2 compatibility
except NameError:
    pass

__doc__ = '''
            PyscesPlot2
            -----------

            PyscesPlot2 is a new graphics susbsystem for PySCeS which will include a
            Unified Plotting Interface which can take advantage of different plotting
            backends via a common user interface.
            '''


class PlotBase(object):
    """
    Abstract class defining the Unified Plotting Interface methods. These methods should
    be overridden and the class extended by interface specific subclasses.
    """

    CommonStyleDefs = {'points': '', 'lines': ''}

    def plot(self, data, x, y, title='', format=''):
        """
        Plot a single line data[y] vs data[x] where:

        - *data* the data array
        - *x* x column index
        - *y* y column index
        - *title* is the line key
        - *format* is the XXX format string (default='')

        Format can also be the *CommonStyle* 'lines' or 'points'
        """
        raise NotImplementedError

    def plotLines(self, data, x, y=[], titles=[], formats=['']):
        """
        Plot a multiple lines data[y1, y2, ] vs data[x] where:

        - *data* the data array
        - *x* x column index
        - *y* is a list of line indexes, if empty all of y not including x is plotted
        - *titles* is a list of line keys, if empty Line1,Line2,Line3 is used
        - *formats* is a list (per line) of XXX format strings.

        If *formats* only contains a single item, this format is used for all lines
        and can also be the *CommonStyle* 'lines' or 'points'.
        """
        raise NotImplementedError

    def splot(self, data, x, y, z, titles='', format=''):
        """
        Plot a surface data[z] vs data[y] vs data[x] where:

        - *data* the data array
        - *x* x column index
        - *y* y column index
        - *z* z column index
        - *title* is the surface key
        - *format* is the XXX format string (default='')

        Format can also be the *CommonStyle* 'lines' or 'points'.
        """
        raise NotImplementedError

    def splotSurfaces(self, data, x, y, z=[], titles=[], formats=['']):
        """
        Plot data[z1, z2, ] vs data[y] vs data[x] where:

        - *data* the data array
        - *x* x column index
        - *y* y column index
        - *z* list of z column indexes, if empty all of z not including x, y are plotted
        - *titles* is a list of surface keys, if empty Surf1, Surf2, Surf3 is used
        - *formats* is a list (per line) of XXX format strings (default='').

        If *formats* only contains a single item, this format is used for all surfaces
        and can also be the *CommonStyle* 'lines' or 'points'.
        """
        raise NotImplementedError

    def setLogScale(self, axis):
        """
        Set axis to logscale where:

        - *axis* = x, y, z, xy, xz, yz, zyx
        """
        raise NotImplementedError

    def setNoLogScale(self, axis):
        """
        Set axis to a linear scale where:

        - *axis* = x, y, z, xy, xz, yz, zyx
        """
        raise NotImplementedError

    def setRange(self, axis, min=None, max=None):
        """
        Set axis range where

        - *axis* = x, y, z, xy, xz, yz, zyx
        - *min* = range(s) lower bound (default=None) autoscale
        - *max* = range(s) upper bound (default=None) autoscale
        """
        raise NotImplementedError

    def setGrid(self, value):
        """
        Display or remove graph grid.

        - *value* (boolean) True (on) or False (off)
        """
        raise NotImplementedError

    def setGraphTitle(self, title='PySCeS Plot'):
        """
        Set the graph title, unset if title=None

        - *title* (string, default='PySCeS Plot') the graph title
        """
        raise NotImplementedError

    def setKey(self, value=False):
        """
        Enable or disable the current plot key, no arguments removes key.

        - *value* boolean (default = False)
        """
        raise NotImplementedError

    def setAxisLabel(self, axis, label=''):
        """
        Set the axis label:

        - *axis* = x, y, z, xy, xz, yz, zyx
        - *label* = string (default='')

        Called with only the axis argument clears the axis label.
        """
        raise NotImplementedError

    def save(self, name, directory=None, dfmt='%.8e'):
        """
        Save the plot data and (optionally) XXX format file

        - *filename* the filename
        - *directory* optional (default = current working directory)
        - *dfmt* the data format string (default='%.8e')
        """
        raise NotImplementedError

    def export(self, name, directory=None, outtype='png'):
        """
        Export the current plot as a <format> file.

        - *filename* the filename
        - *directory* optional (default = current working directory)
        - *outtype* the file format (default='png').

        Currently only PNG is guaranteed to be available in all interfaces.
        """
        raise NotImplementedError

    def axisInputStringToList(self, input):
        """
        Extracts axis information from a string input, returns a boolean triple
        representing (x=True/False, y=True/False, z=True/False).

        - *input* the input string
        """
        out = [None, None, None]
        input = input.lower()
        if 'x' in input:
            out[0] = 'x'
        if 'y' in input:
            out[1] = 'y'
        if 'z' in input:
            out[2] = 'z'
        return out

    def wait(self, seconds=3):
        """
        Wait *seconds* (default = 3) or until enter is pressed (seconds = -1)
        """
        if seconds == -1:
            input("\nPress <enter> to continue ...\n")
        else:
            time.sleep(seconds)


class GnuPlotUPI(PlotBase):
    """
    PySCeS/GnuPlot is reborn, leaner and meaner than ever before. This class enables
    plotting with GnuPlot via a subprocess link:

    - *work_dir* optional argument setting directory for dat file(s)
    - *gnuplot_dir* optional argument specifying the location of pgnuplot.exe (win32) or gnuplot

    GnuPlot backend to the Unified Plotting Interface.
    """

    __G_SUBPROC__ = None
    __ECHO__ = False
    __DATA_FILE_NAME__ = '_gnuplot.dat'
    __WORK_DIR__ = None
    __DATA_FILE_PATH__ = None
    DATF_FORMAT = '%.8e'
    __LAST_COMMAND__ = ''
    __MULTIPLOT__ = False
    PAUSE_TIME = 0.1
    __GNUPLOT_EXE_PATH__ = None
    __DISP_TERMINAL__ = None
    ##  MULTIPLOT_CNTR = 0
    ##  MULTIPLOT_MAX = 3
    ##  datFbuffsize = 5
    ##  datFcache = None

    CommonStyleDefs = {'points': 'w p', 'lines': 'w l'}

    Terminals = {
        'x11': '',
        'windows': '',
        'png': 'medium size 800,600'
        # 'xffffff x000000 x404040 xff0000 xffa500 x66cdaa xcdb5cd xadd8e6 x0000ff xdda0dd x9500d3'
    }

    def __init__(self, work_dir=None, gnuplot_dir=None):
        if gnuplot_dir != None or not os.path.exists(gnuplot_dir):
            GNUPLOT_PATH = gnuplot_dir
        else:
            GNUPLOT_PATH = ''
        if os.sys.platform == 'win32':
            self.__GNUPLOT_EXE_PATH__ = os.path.join(GNUPLOT_PATH, 'pgnuplot.exe')
            self.__DISP_TERMINAL__ = 'windows'
        else:
            self.__GNUPLOT_EXE_PATH__ = os.path.join(GNUPLOT_PATH, 'gnuplot')
            self.__DISP_TERMINAL__ = 'x11'
        try:
            self.__G_SUBPROC__ = subprocess.Popen(
                self.__GNUPLOT_EXE_PATH__, stdin=subprocess.PIPE
            )
        except Exception as ex:
            print('\nGnuPlot load failure\n')
            print(ex)
        if work_dir == None or not os.path.exists(work_dir):
            work_dir = os.getcwd()
        self.__DATA_FILE_PATH__ = os.path.join(work_dir, self.__DATA_FILE_NAME__)
        self.__WORK_DIR__ = work_dir

    def set(self, key, value=''):
        """
        Send *set <key>* or optionally *set <key> <value>* to GnuPlot.
        """
        self.g_write("set %s %s" % (key, value))

    def unset(self, key, value=''):
        """
        Send *unset <key>* or optionally *unset <key> <value>* to GnuPlot.
        """
        self.g_write("unset %s %s" % (key, value))

    def g_write(self, cmd):
        """
        Write a command to the GnuPlot interpreter

        - *cmd* the GnuPlot command
        """
        cmd = cmd.replace('\\', '/')
        self.last_command = cmd
        if self.__ECHO__:
            print('SENT: %s' % cmd)

        self.__G_SUBPROC__.stdin.write("""%s\n""" % cmd)

    def g_file_write_array(self, arr, dfmt=None):
        """
        Write a normal (2D) dataset to temp file. Dumps the array to file
        using the format:

        - *arr* the array (r>0, c>1)
        - *fmt* default '%.8e'
        """
        if dfmt == None:
            dfmt = self.DATF_FORMAT
        numpy.savetxt(self.__DATA_FILE_PATH__, arr, fmt=dfmt, delimiter=' ')
        if self.__ECHO__:
            print('WROTE: %s' % self.__DATA_FILE_PATH__)
        self.g_pause()  # multiplot protection

    def g_file_write_array3D(self, arr, yaxis=1, dfmt=None):
        """
        Write a GnuPlot format 3D dataset. The *yaxis* argument
        specifies the column that should be used to split the dataset
        into GnuPlot slices.

        - *arr* the array (r>1, c>2)
        - *fmt* default '%.8e'
        - *yaxis* default 1
        """
        outlist = []
        set_idx = 0
        curr_y = arr[0, yaxis]
        for r in range(arr.shape[0]):
            if arr[r, yaxis] != curr_y:
                outlist.append(arr[set_idx:r, :])
                curr_y = arr[r, yaxis]
                set_idx = r
        if set_idx < arr.shape[0]:
            outlist.append(arr[set_idx:, :])

        if dfmt == None:
            dfmt = self.DATF_FORMAT
        if len(outlist) <= 1:
            self.g_file_write_array(arr, dfmt=dfmt)
        else:
            F = open(self.__DATA_FILE_PATH__, 'w+')
            for d in range(len(outlist)):
                numpy.savetxt(F, outlist.pop(0), fmt=dfmt)
                F.write(' \n')
            F.flush()
            F.close()
            if self.__ECHO__:
                print('WROTE: %s' % self.__DATA_FILE_PATH__)
        del outlist
        self.g_pause()

    def g_pause(self):
        """
        A small pause defined by *self.PAUSE_TIME* (multiplied by 2 when in multiplot).
        """
        if self.__MULTIPLOT__:
            time.sleep(self.PAUSE_TIME * 2.0)
        else:
            time.sleep(self.PAUSE_TIME)

    def plot(self, data, x, y, title='', format='w l'):
        """
        Plot a single line data[y] vs data[x] where:

        - *data* the data array
        - *x* x column index
        - *y* y column index
        - *title* is the line key
        - *format* is the GnuPlot format string (default='w l')

        Format can also be the *CommonStyle* 'lines' or 'points'.
        """
        if format in list(self.CommonStyleDefs.keys()):
            format = self.CommonStyleDefs[format]
        if title == '':
            title = []
        else:
            title = ['X', title]
        if format == '':
            format = 'w l'
        self.plotLines(data, x, [y], title, [format] * data.shape[1])

    def plotLines(self, data, x, y=[], titles=[], formats=['w l']):
        """
        Plot a multiple lines data[y1, y2, ] vs data[x] where:

        - *data* the data array
        - *x* x column index
        - *y* is a list of line indexes, if empty all of y not including x is plotted
        - *titles* is a list of line keys if empty Line1, Line2, Line3 is used
        - *formats* is a list (per line) of GnuPlot format strings (default='w l').

        If *formats* only contains a single item, this format is used for all lines
        and can also be the *CommonStyle* 'lines' or 'points'.
        """

        self.g_file_write_array(data, dfmt=self.DATF_FORMAT)
        if len(y) == 0:
            y = list(range(data.shape[1]))
            y.pop(x)
        if len(formats) != 1 and len(formats) != data.shape[1]:
            print(
                "len(titles) must be one or equal the number data columns (%s) using: %s"
                % (data.shape[1], formats[-1])
            )
            formats = [formats[-1]]
        elif formats[0] == '' or formats[0] == None:
            formats = ['w l']
        if len(formats) == 1 and data.shape[1] > 1:
            formats = data.shape[1] * formats
        cFormats = list(self.CommonStyleDefs.keys())
        for f in range(len(formats)):
            if formats[f] in cFormats:
                formats[f] = self.CommonStyleDefs[formats[f]]
        if len(titles) == 0:
            titles = ['col%s' % (l) for l in range(data.shape[1])]

        cmd = '''plot '''
        for yi in y:
            cmd += '''"%s" u %s:%s t "%s" %s, ''' % (
                self.__DATA_FILE_PATH__,
                x + 1,
                yi + 1,
                titles[yi],
                formats[yi],
            )
        cmd = cmd[:-2]
        self.g_write(cmd)
        self.setKey(True)

    def splot(self, data, x, y, z, titles='', format='w l'):
        """
        Plot a surface data[z] vs data[y] vs data[x] where:

        - *data* the data array
        - *x* x column index
        - *y* y column index
        - *z* z column index
        - *titles* is the surface key
        - *format* is the GnuPlot format string (default='w l')

        Format can also be the *CommonStyle* 'lines' or 'points'.
        """
        if format in list(self.CommonStyleDefs.keys()):
            format = self.CommonStyleDefs[format]
        if titles == '':
            titles = []
        else:
            assert (
                len(titles) == data.shape[1]
            ), '\nTitle list must match number of columns'
        if format == '':
            format = 'w l'
        self.splotSurfaces(data, x, y, [z], titles, [format] * data.shape[1])

    def splotSurfaces(self, data, x, y, z=[], titles=[], formats=['w l']):
        """
        Plot data[z1, z2, ] vs data[y] vs data[x] where:

        - *data* the data array
        - *x* x column index
        - *y* y column index
        - *z* list of z column indexes, if empty all of z not including x, y are plotted
        - *titles* is a list of surface keys, if empty Surf1, Surf2, Surf3 is used
        - *formats* is a list (per line) of GnuPlot format strings (default='w l').

        If *formats* only contains a single item, this format is used for all surface and
        can also be the *CommonStyle* 'lines' or 'points'.
        """
        self.g_file_write_array3D(data, yaxis=y, dfmt=self.DATF_FORMAT)

        if len(z) == 0:
            z = list(range(data.shape[1]))
        if len(formats) != 1 and len(formats) != data.shape[1]:
            print(
                "len(titles) must be one or equal the number data columns (%s) using: %s"
                % (data.shape[1], formats[-1])
            )
            formats = [formats[-1]]
        elif formats[0] == '' or formats[0] == None:
            formats = ['w l']
        if len(formats) == 1 and data.shape[1] > 1:
            formats = data.shape[1] * formats
        cFormats = list(self.CommonStyleDefs.keys())
        for f in range(len(formats)):
            if formats[f] in cFormats:
                formats[f] = self.CommonStyleDefs[formats[f]]
        if len(titles) == 0:
            titles = ['col%s' % (s + 1) for s in range(data.shape[1])]

        cmd = '''splot '''

        for zi in z:
            if zi not in (x, y):
                cmd += '''"%s" u %s:%s:%s t "%s" %s, ''' % (
                    self.__DATA_FILE_PATH__,
                    x + 1,
                    y + 1,
                    zi + 1,
                    titles[zi],
                    formats[zi],
                )
        cmd = cmd[:-2]
        print(cmd)
        self.g_write(cmd)
        self.setKey(True)

    def save(self, name, directory=None, dfmt=None):
        """
        Save the last plot as a GnuPlot file *name*.plt which references
        *name*.dat.

        - *name* the name of the GnuPlot plt and and datafile
        - *directory* (optional) the directory to use (defaults to working directory)
        - *dfmt* is ignored and uses the value of self.DATF_FORMAT
        """
        if directory != None:
            out_n = os.path.join(directory, name)
        else:
            out_n = os.path.join(self.__WORK_DIR__, name)
        self.g_write('save "%s.plt"' % out_n)
        shutil.copy(
            os.path.join(self.__WORK_DIR__, self.__DATA_FILE_NAME__), "%s.dat" % out_n
        )
        F = open('%s.plt' % out_n, 'r')
        fnew = F.read().replace('_gnuplot.dat', '%s.dat' % name)
        fnew = fnew.replace(
            'noequal_axes', ''
        )  # fixes the "noequal_axes" bug in gnuplot save
        fnew = fnew.replace(
            '#!/gnuplot',
            '#!/gnuplot\n#\n# Plot created using PySCeS %s (http://pysces.sourceforge.net)'
            % __version__,
        )  # fixes the "noequal_axes" bug in gnuplot save
        F.close()
        F = open('%s.plt' % out_n, 'w')
        F.write(fnew)
        F.flush()
        F.close()

    def replot(self):
        """
        Replot the current GnuPlot plot
        """
        self.g_write('replot')
        self.g_pause()

    def replotAndWait(self, seconds=0.5):
        """
        Replot the current GnuPlot plot and wait default (*seconds* = 0.5)
        or until enter is pressed (*seconds* = -1)
        """
        self.replot()
        self.wait(seconds)

    def setLogScale(self, axis):
        """
        Set axis to logscale where:

        - *axis* = x, y, z, xy, xz, yz, zyx
        """
        axout = self.axisInputStringToList(axis)
        axstr = ''
        for s in axout:
            if s != None:
                axstr += s
        self.set('logscale', axstr)

    def setNoLogScale(self, axis):
        """
        Set axis to a linear scale where:

        - *axis* = x, y, z, xy, xz, yz, zyx
        """
        axout = self.axisInputStringToList(axis)
        axstr = ''
        for s in axout:
            if s != None:
                axstr += s
        self.unset('logscale', axstr)

    def setRange(self, axis, min=None, max=None):
        """
        Set axis range where:

        - *axis* = x, y, z, xy, xz, yz, zyx
        - *min* = range(s) lower bound (default=None) autoscale
        - *max* = range(s) upper bound (default=None) autoscale

        If only the *axis* argument is provided, GnuPlot will autoscale the
        ranges to the data.
        """
        axout = self.axisInputStringToList(axis)
        if min == None:
            min = '*'
        else:
            min = '%s' % min
        if max == None:
            max = '*'
        else:
            max = '%s' % max
        for ax in axout:
            if ax != None:
                self.set('%srange' % ax, '[ %s : %s ]' % (min, max))

    def setGrid(self, value):
        """
        Display or remove graph grid.

        - *value* (boolean) True (on) or False (off)
        """
        if value:
            self.set('grid')
        else:
            self.unset('grid')

    def setGraphTitle(self, title='PySCeS Plot'):
        """
        Set the graph title, unset if title argument is None

        - *title* (string, default='PySCeS Plot') the graph title
        """
        if title == None:
            self.unset('title')
        else:
            self.set('title', '\"%s\"' % title)

    def setKey(self, value=False):
        """
        Enable or disable the current plot key, no arguments removes key.

        - *value* boolean (default = False)
        """
        if value:
            self.set('key')
        else:
            self.unset('key')

    def setAxisLabel(self, axis, label=''):
        """
        Set the axis label:

        - *axis* = x, y, z, xy, xz, yz, zyx
        - *label* = string (default='')

        Called with only the axis argument clears the axis label.
        """
        axout = self.axisInputStringToList(axis)
        for ax in axout:
            if ax != None:
                self.set('%slabel \"%s\"' % (ax, label))

    def export(self, name, directory=None, outtype='png'):
        """
        Export the current plot as a <format> file.

        - *filename* the filename
        - *directory* optional (default = current working directory)
        - *outtype* the file format (default='png').

        Currently only PNG is guaranteed to be available in all interfaces.
        """
        imgPath = None
        if outtype.lower() == 'png':
            if name[-4:] != '.%s' % outtype:
                name += '.%s' % outtype
            if directory != None and os.path.exists(directory):
                imgPath = os.path.join(directory, name)
            else:
                imgPath = os.path.join(os.getcwd(), name)

        if imgPath != None and outtype in list(self.Terminals.keys()):
            self.set('output', '\"%s\"' % imgPath)
            self.setTerminal(outtype, self.Terminals[outtype])
            self.replot()
            self.set('output')
            self.setTerminal(
                self.__DISP_TERMINAL__, self.Terminals[self.__DISP_TERMINAL__]
            )
            print('Image exported as \"%s\"' % imgPath)
        else:
            print('Image not exported as type \"%s\"' % outtype)

    def setDataFileNumberFormat(self, format='%.8e'):
        """
        Sets the format string for data written to file

        - *format* format string (default='%.8e')
        """
        print('GnuPlot array data will be written as %s' % format % 0.123456789)
        self.DATF_FORMAT = format

    def setTerminal(self, name, options=''):
        """
        Sets the terminal, gnuplot: set terminal *name* *options*
        """
        self.set('terminal', '%s %s' % (name, options))

    def setMultiplot(self):
        """
        Begin a multiplot session
        """
        self.__MULTIPLOT__ = True
        self.set("multiplot")

    def unsetMultiplot(self):
        """
        End a multiplot session.
        """
        self.__MULTIPLOT__ = False
        self.unset("multiplot")

    def setSize(self, width=1.0, height=1.0):
        """
        Set the size of the next plot relative to the GnuPlot canvas (e.g. screen)
        size which is defined to be 1. For example if ``width = height = 0.5`` the
        plot is 1/4 the size of the viewable canvas. If no arguments are supplied
        reset size to 1,1.

        - *width* of next plot (default = 1.0)
        - *height* of next plot (default = 1.0)
        """
        self.set("size", "%s,%s" % (width, height))

    def setOrigin(self, xpos=0, ypos=0):
        """
        Set the origin (lower left corner) of the next plot. Uses GnuPlot
        screen coordinates. If no arguments are supplied reset origin to 0,0.

        - *xpos* of next plot (default = 0)
        - *ypos* of next plot (default = 0)
        """
        self.set("origin", "%s,%s" % (xpos, ypos))

    def setSizeAndOrigin(self, width=1, height=1, xpos=0, ypos=0):
        """
        Set the size and origin of the next plot. If no arguments
        are supplied, reset the size to 1,1 and origin to 0.0

        - *width* of next plot (default = 1.0)
        - *height* of next plot (default = 1.0)
        - *xpos* of next plot (default = 0)
        - *ypos* of next plot (default = 0)
        """
        self.setSize(width, height)
        self.setOrigin(xpos, ypos)


class MatplotlibUPI(PlotBase):
    """
    Refactored Matplotlib backend to the Unified Plotting Interface

    - *work_dir* (optional) working directory
    """

    __MODE_INTERACTIVE__ = True
    __MODE_NEW_FIGURE__ = True
    __FIGURE_GENERATOR__ = None
    __CURRENT_FIGURE__ = None
    __MODE_HOLD__ = False
    MAX_OPEN_WINDOWS = 10
    __ARRAY_DATA__ = None
    __WORK_DIR__ = None
    pyplot = None
    # __ENABLE_HTML5__ = False

    CommonStyleDefs = {'points': 'o', 'lines': '-'}
    __BACKENDS__ = (
        'gtk',
        'gtkagg',
        'gtkcairo',
        'fltkagg',
        'macosx',
        'qtagg',
        'qt4agg',
        'tkagg',
        'wx',
        'wxagg',
        'cocoaagg',
        'agg',
        'nbagg',
        'cairo',
        'emf',
        'gdk',
        'pdf',
        'ps',
        'svg',
        'template',
    )

    __INTERACTIVE_BACKENDS__ = [
        'gtk',
        'gtkagg',
        'gtkcairo',
        'fltkagg',
        'macosx',
        'qtagg',
        'qt4agg',
        'tkagg',
        'wx',
        'wxagg',
        'cocoaagg',
        'nbagg',
    ]

    __NON_INTERACTIVE_BACKENDS__ = [
        'agg',
        'pdf',
        'ps',
        'svg',
        ]

    __BACKEND__ = None

    def __init__(self, work_dir=None, backend=None):
        if work_dir != None and os.path.exists(work_dir):
            self.__WORK_DIR__ = work_dir
        else:
            self.__WORK_DIR__ = os.getcwd()
        try:
            import matplotlib

            if self.isnotebook():
                backend = 'nbAgg'
                import pysces
                pysces.__MATPLOTLIB_BACKEND__ = 'nbAgg'

            if backend.lower() in self.__INTERACTIVE_BACKENDS__:
                matplotlib.use(backend)
                self.__BACKEND__ = backend
                if not __SILENT_START__:
                    print(('Matplotlib backend set to: \"{}\"'.format(backend)))
            elif backend.lower() in self.__NON_INTERACTIVE_BACKENDS__:
                matplotlib.use(backend)
                self.__BACKEND__ = backend
                if not __SILENT_START__:
                    print(('Matplotlib backend set to non-interactive backend: \"{}\"'.format(backend)))
            elif backend.lower() == 'native':
                    if not __SILENT_START__:
                        print(
                            (
                                'Using natively configured Matplotlib backend: \"{}\"'.format(
                                    matplotlib.get_backend()
                                )
                            )
                        )
                    self.__BACKEND__ = matplotlib.get_backend()
            else:
                matplotlib.use('TkAgg')
                self.__BACKEND__ = 'TkAgg'
                if not __SILENT_START__:
                    print(
                        (
                            'Matplotlib \"{}\" backend not set, defaulting to: \"{}\"'.format(
                                backend, 'TkAgg'
                            )
                        )
                    )

            # if self.__ENABLE_HTML5__:
            # matplotlib.use('module://mplh5canvas.backend_h5canvas') # HTML5
            # else:
            # matplotlib.use('TKagg', warn=False)
        except Exception as ex:
            if not __SILENT_START__:
                print(ex)
                print(
                    "\nPySCeS defaults to matplotlib's TKagg backend if not specified \
                         in the user configuration file, set \"matplotlib_backend = <backend>\" "
                )

        from matplotlib import pyplot
        from matplotlib import pylab

        ##  self.pyplot = pyplot
        self.pyplot = pylab
        if self.__MODE_INTERACTIVE__:
            self.pyplot.ion()
        self._setNewFigureGenerator(self.MAX_OPEN_WINDOWS)

    def isnotebook(self):
        try:
            shell = get_ipython().__class__.__name__
            if shell == 'ZMQInteractiveShell':
                return True  # Jupyter notebook or qtconsole
            elif shell == 'TerminalInteractiveShell':
                return False  # Terminal running IPython
            else:
                return False  # Other type (?)
        except NameError:
            return False  # Probably standard Python interpreter

    def _setNewFigureGenerator(self, num):
        """
        Create a figure_generator with range [1,num+1]
        """
        self.__FIGURE_GENERATOR__ = itertools.cycle(list(range(1, num + 1)))
        self.MAX_OPEN_WINDOWS = num

    def _setActiveFigure(self, fignum):
        self.__CURRENT_FIGURE__ = self.pyplot.figure(fignum)

    def _setNewFigure(self):
        self._setActiveFigure(next(self.__FIGURE_GENERATOR__))

    def closeAll(self):
        """
        Close all open matplotlib figures.
        """
        for x in range(self.MAX_OPEN_WINDOWS):
            self.pyplot.close()

    def hold(self, hold=False):
        """
        Enable plot holding where each new graph is plotted on top of
        the previous one.

        - *hold* boolean (default = False)
        """
        if hold:
            self.__MODE_NEW_FIGURE__ = False
            self.__MODE_HOLD__ = True
        else:
            self.__MODE_NEW_FIGURE__ = True
            self.__MODE_HOLD__ = False

    def plot(self, data, x, y, title='', format='-'):
        """
        Plot a single line data[y] vs data[x] where:

        - *data* the data array
        - *x* x column index
        - *y* y column index
        - *title* is the line key
        - *format* is the Matplotlib format string (default='-')

        Format can also be the *CommonStyle* 'lines' or 'points'.
        """
        if format in list(self.CommonStyleDefs.keys()):
            format = self.CommonStyleDefs[format]
        if title == '':
            title = []
        else:
            title = data.shape[1] * [title]
        # print(title)

        self.plotLines(data, x, [y], title, [format] * data.shape[1])

    def plotLines(self, data, x, y=[], titles=[], formats=['-']):
        """
        Plot a multiple lines data[y1, y2, ] vs data[x] where:

        - *data* the data array
        - *x* x column index
        - *y* is a list of line indexes
        - *titles* is a list of line keys
        - *formats* is a list (per line) of Matplotlib format strings.

        If *formats* only contains a single item, this format is used for all lines
        and can also be the *CommonStyle* 'lines' or 'points'.
        """
        self.__ARRAY_DATA__ = data
        if self.__MODE_NEW_FIGURE__:
            self._setNewFigure()
        if not self.__MODE_HOLD__:
            self.pyplot.clf()
        if len(y) == 0:
            y = list(range(self.__ARRAY_DATA__.shape[1]))
            y.pop(x)

        if len(formats) != 1 and len(formats) != data.shape[1]:
            print(
                "len(titles) must be one or equal the number data columns (%s) using: \"%s\""
                % (data.shape[1], formats[-1])
            )
            formats = [copy.copy(formats[-1])]
        elif formats[0] == '' or formats[0] == None:
            formats = ['-']
        if len(formats) == 1 and data.shape[1] > 1:
            formats = data.shape[1] * formats
        cFormats = list(self.CommonStyleDefs.keys())
        for f in range(len(formats)):
            if formats[f] in cFormats:
                formats[f] = self.CommonStyleDefs[formats[f]]
        if len(titles) == 0:
            titles = ['col%s' % (l) for l in range(data.shape[1])]

        self.pyplot.ioff()
        for yi in y:
            # print(self.__ARRAY_DATA__.take([x],axis=1))
            # print(self.__ARRAY_DATA__.take([yi],axis=1))
            self.pyplot.plot(
                self.__ARRAY_DATA__.take([x], axis=1),
                self.__ARRAY_DATA__.take([yi], axis=1),
                formats[yi],
                label=titles[yi],
            )
        if self.__MODE_INTERACTIVE__:
            self.pyplot.ion()
        self.pyplot.legend()
        # if self.__ENABLE_HTML5__:
        # self.pyplot.show()

    def setLogScale(self, axis):
        """
        Set axis to logscale where:

        - *axis* = x, y, z, xy, xz, yz, zyx
        """
        axout = self.axisInputStringToList(axis)
        A = self.pyplot.gca()
        for ax in axout:
            if ax != None and ax in ['x', 'y']:
                getattr(A, 'set_%sscale' % ax)('log')
        if self.__MODE_INTERACTIVE__:
            self.pyplot.draw()

    def setNoLogScale(self, axis):
        """
        Set axis to a linear scale where:

        - *axis* = x, y, z, xy, xz, yz, zyx
        """
        axout = self.axisInputStringToList(axis)
        A = self.pyplot.gca()
        for ax in axout:
            if ax != None and ax in ['x', 'y']:
                getattr(A, 'set_%sscale' % ax)('linear')
        if self.__MODE_INTERACTIVE__:
            self.pyplot.draw()

    def setRange(self, axis, min=None, max=None):
        """
        Set axis range where

        - *axis* = x, y, z, xy, xz, yz, zyx
        - *min* = range(s) lower bound (default=None) autoscale
        - *max* = range(s) upper bound (default=None) autoscale
        """
        axout = self.axisInputStringToList(axis)
        for ax in axout:
            if ax != None and ax in ['x', 'y']:
                getattr(self.pyplot, '%slim' % ax)(min, max)
        ##  if self.__MODE_INTERACTIVE__: self.pyplot.draw()

    def setGrid(self, value):
        """
        Display or remove graph grid.

        - *value* (boolean) True (on) or False (off)
        """
        self.pyplot.grid(value)

    def setGraphTitle(self, title='PySCeS Plot'):
        """
        Set the graph title, unset if title=None

        - *title* (string, default='PySCeS Plot') the graph title
        """
        self.pyplot.title(title)

    def setKey(self, value=False):
        """
        Enable or disable the current plot key, no arguments removes key.

        - *value* boolean (default = False)
        """
        L = self.pyplot.gca().get_legend()
        if value:
            L._visible = True
        else:
            L._visible = False
        if self.__MODE_INTERACTIVE__:
            self.pyplot.draw()

    def setAxisLabel(self, axis, label=''):
        """
        Set the axis label:

        - *axis* = x, y, z, xy, xz, yz, zyx
        - *label* = string (default='')

        Called with only the axis argument clears the axis label.
        """
        axout = self.axisInputStringToList(axis)
        for ax in axout:
            if ax in ['x', 'y']:
                getattr(self.pyplot, '%slabel' % ax)(label)

    def save(self, name, directory=None, dfmt='%.8e'):
        """
        Save the plot data to

        - *filename* the filename
        - *directory* optional (default = current working directory)
        - *dfmt* the data format string (default='%.8e')
        """
        if directory != None:
            out_n = os.path.join(directory, name)
        else:
            out_n = os.path.join(self.__WORK_DIR__, name)
        numpy.savetxt(out_n, self.__ARRAY_DATA__, fmt=dfmt, delimiter=' ')
        print('Data saved as \"%s\"' % out_n)

    def export(self, name, directory=None, outtype='png'):
        """
        Export the current plot as a <format> file.

        - *filename* the filename
        - *directory* optional (default = current working directory)
        - *outtype* the file format (default='png').

        Currently only PNG is guaranteed to be available in all interfaces.
        """
        imgPath = None
        if self.__BACKEND__ in self.__NON_INTERACTIVE_BACKENDS__:
            terminals = self.__NON_INTERACTIVE_BACKENDS__
            outtype = self.__BACKEND__
            print('Using non-interactice terminal as outtype: \"{}\"'.format(outtype.lower()))
        else:
            terminals = ['png']
        if outtype.lower() in terminals:
            if name[-4:] != '.%s' % outtype:
                name += '.%s' % outtype
            if directory != None and os.path.exists(directory):
                imgPath = os.path.join(directory, name)
            else:
                imgPath = os.path.join(os.getcwd(), name)
            self.pyplot.savefig(imgPath)
            print('Image exported as \"%s\"' % imgPath)
        else:
            print('Image not exported as type \"%s\"' % outtype)

    def setLineWidth(self, width=1):
        """
        Sets the line width for current axis

        - *width* the line width
        """
        [l.set_linewidth(width) for l in self.pyplot.gca().get_lines()]
        if self.__MODE_INTERACTIVE__:
            self.pyplot.draw()


class PyscesUPI(PlotBase):
    """
    This is the frontend to the PySCeS Unified Plotting Interface (pysces.plt.*) that
    allows one to specify which backend should be used to plot when a
    UPI method is called. More than one interface can be active at the same time and
    so far the Matplotlib and GnuPlot backends are available for use.

    This is an experiment which must be refactored into a more general
    way of doing things. Basically, I want an instance of the abstract plotting
    class which will plot to one, any or all currently available backends.
    If anybody has an idea how I can generate this class automatically please let
    me know ;-)
    """

    g = None
    m = None
    __USE_GNUPLOT__ = False
    __USE_MATPLOTLIB__ = False
    __GINTERFACES__ = ('matplotlib', 'gnuplot')

    def p_setInterface(self, name, instance):
        """
        Add an interface to the backend selector

        - *name* the interface name currently one of ['matplotlib','gnuplot']
        - *instance* an instance of a PlotBase derived (UPI) interface
        """
        if name in self.__GINTERFACES__:
            self.__setattr__('__USE_%s__' % name.upper(), True)
            self.__setattr__('%s' % name[0], instance)

    def p_deactivateInterface(self, interface):
        """
        Deactivate the interface. This does not delete the interface and
        it is possible to reactivate the deactivated interface with **p_activateInterface**.

        - *interface* one of ['matplotlib','gnuplot']
        """
        if interface in self.__GINTERFACES__:
            self.__setattr__('__USE_%s__' % interface.upper(), False)

    def p_activateInterface(self, interface):
        """
        Activate an interface that has been set with **p_setInterface()** but deactivated with
        **p_deactivateInterface**

        - *interface* one of ['matplotlib','gnuplot']
        """
        if (
            interface in self.__GINTERFACES__
            and self.__getattribute__('%s' % interface[0]) != None
        ):
            self.__setattr__('__USE_%s__' % interface.upper(), True)
        else:
            print('Cannot activate \"%s\"interface' % interface)

    def __gselect__(self, func, *args, **kwargs):
        """
        This methods uses plots a UPI method to *any* activated plotting
        interface
        """
        # avoids fishy nested args bug (not really relevant anymore but here just in case)
        if self.__USE_GNUPLOT__ and self.__USE_MATPLOTLIB__:
            # args2 = tuple([copy.deepcopy(a) for a in args])
            ##  args2 = copy.deepcopy(args)
            ##  kwargs2 = copy.deepcopy(kwargs)
            args2 = args
            kwargs2 = kwargs
        else:
            args2 = args
            kwargs2 = kwargs
        if self.__USE_MATPLOTLIB__:
            try:
                self.m.__getattribute__(func)(*args, **kwargs)
            except NotImplementedError:
                print('\"%s\" method not available using Matplotlib' % func)

        if self.__USE_GNUPLOT__:
            try:
                self.g.__getattribute__(func)(*args2, **kwargs2)
            except NotImplementedError:
                print('\"%s\" method not available using GnuPlot ' % func)
        del args, kwargs, args2, kwargs2

    def plot(self, data, x, y, title='', format=''):
        """
        Plot a single line data[y] vs data[x] where:

        - *data* the data array
        - *x* x column index
        - *y* y column index
        - *title* is the line key
        - *format* is the backend format string (default='')
        """
        self.__gselect__('plot', data, x, y, title, format)

    def plotLines(self, data, x, y=[], titles=[], formats=['']):
        """
        Plot a multiple lines data[y1, y2, ] vs data[x] where:

        - *data* the data array
        - *x* x column index
        - *y* is a list of line indexes, if empty all of y not including x is plotted
        - *titles* is a list of line keys, if empty Line1,Line2,Line3 is used
        - *formats* is a list (per line) of XXX format strings.

        If *formats* only contains a single item, this format is used for all lines.
        """
        self.__gselect__('plotLines', data, x, y, titles, formats)

    def splot(self, data, x, y, z, titles='', format=''):
        """
        Plot a surface data[z] vs data[y] vs data[x] where:

        - *data* the data array
        - *x* x column index
        - *y* y column index
        - *z* z column index
        - *titles* is a list of surface keys whose len matches data columns
        - *format* is the XXX format string (default='')
        """
        self.__gselect__('splot', data, x, y, z, titles, format)

    def splotSurfaces(self, data, x, y, z=[], titles=[], formats=['']):
        """
        Plot data[z1, z2, ] vs data[y] vs data[x] where:

        - *data* the data array
        - *x* x column index
        - *y* y column index
        - *z* list of z column indexes, if empty all of z not including x, y are plotted
        - *titles* is a list of surface keys, if empty Surf1, Surf2, Surf3 is used
        - *formats* is a list (per line) of XXX format strings (default='').

        If *formats* only contains a single item, this format is used for all surfaces.
        """
        self.__gselect__('splotSurfaces', data, x, y, z, titles, formats)

    def setLogScale(self, axis):
        """
        Set axis to logscale where:

        - *axis* = x, y, z, xy, xz, yz, zyx
        """
        self.__gselect__('setLogScale', axis)

    def setNoLogScale(self, axis):
        """
        Set axis to a linear scale where:

        - *axis* = x, y, z, xy, xz, yz, zyx
        """
        self.__gselect__('setNoLogScale', axis)

    def setRange(self, axis, min=None, max=None):
        """
        Set axis range where

        - *axis* = x, y, z, xy, xz, yz, zyx
        - *min* = range(s) lower bound (default=None) autoscale
        - *max* = range(s) upper bound (default=None) autoscale
        """
        self.__gselect__('setRange', axis, min, max)

    def setGrid(self, value):
        """
        Display or remove graph grid.

        - *value* (boolean) True (on) or False (off)
        """
        self.__gselect__('setGrid', value)

    def setGraphTitle(self, title='PySCeS Plot'):
        """
        Set the graph title, unset if title=None

        - *title* (string, default='PySCeS Plot') the graph title
        """
        self.__gselect__('setGraphTitle', title)

    def setKey(self, value=False):
        """
        Enable or disable the current plot key, no arguments removes key.

        - *value* boolean (default = False)
        """
        self.__gselect__('setKey', value)

    def setAxisLabel(self, axis, label=''):
        """
        Set the axis label:

        - *axis* = x, y, z, xy, xz, yz, zyx
        - *label* = string (default=None)

        Called with only the axis argument clears the axis label.
        """
        self.__gselect__('setAxisLabel', axis, label)

    def save(self, name, directory=None, dfmt='%.8e'):
        """
        Save the plot data and (optionally) XXX format file

        - *filename* the filename
        - *directory* optional (default = current working directory)
        - *dfmt* the data format string (default='%.8e')
        """
        self.__gselect__('save', name, directory, dfmt)

    def export(self, name, directory=None, outtype='png'):
        """
        Export the current plot as a <format> file.

        - *filename* the filename
        - *directory* optional (default = current working directory)
        - *outtype* the file format (default='png').

        Currently only PNG is guaranteed to be available in all interfaces.
        """
        self.__gselect__('export', name, directory, outtype)

    def replot(self):
        """
        Replot the current figure for all active interfaces
        """
        if self.__USE_GNUPLOT__:
            self.g.replot()
        if self.__USE_MATPLOTLIB__:
            self.m.pyplot.draw()

    def closeAll(self):
        """Close all active Matplolib figures"""
        if self.__USE_MATPLOTLIB__:
            self.m.closeAll()


class FIFOBuffer:
    """
    Simple fixed size FIFO buffer.
    """

    def __init__(self, size):
        self.data = [None for i in range(size)]

    def add(self, x):
        self.data.pop(0)
        self.data.append(x)

    def get(self):
        return self.data


# The following code will be adapted used in future for higher level
# graphics operations.
'''
class LineObj2(object):
    """
    Class describing a "graph line" or series of data.

    - **data** a ndarray column of data shape = (r, 1)
    - **label** a string label describing the column
    - **\*\*kwargs** (optional) key=value pairs of property descriptors

    Properties are stored as LineObj2 attributes
    """
    data = None
    label = None
    type = None
    axis = None
    IS_SELECTED = False

    __properties__ = ['index', 'linewidth', 'pointsize', 'colour', 'style',\
                    'logscale', 'visible']

    def __init__(self, data, label, **kwargs):
        self.axis = AxisObj2()
        self.setData(data)
        self.label = label
        self.type = data.dtype.char
        self.setProperties(**kwargs)

    def setProperties(self, **kwargs):
        for k in kwargs.keys():
            if k in self.__properties__:
                setattr(self, k, kwargs[k])

    def getData(self):
        return self.data

    def setData(self, data):
        self.axis.setRange(min(data), max(data))
        self.data = data

    def getLabel(self):
        return self.label

    def getType(self):
        return self.type

    def setSelected(self):
        self.IS_SELECTED = True

    def unsetSelected(self):
        self.IS_SELECTED = False

    def getProperties(self):
        out = {}
        for p in self.__properties__:
            if hasattr(self, p):
                out.update({p : getattr(self, p)})
        return out

    def unsetProperty(self, prop):
        if hasattr(self, prop):
            self.__delattr__(prop)

    def unsetAllProperties(self):
        for p in self.__properties__:
             self.unsetProperty(p)


class AxisObj2(object):
    """
    Class holding axis information (usually found as the *axis* attribute of
    a LineObj2):

    - **IS_AXIS** boolean
    - **IS_LOG** boolean
    - **axis** axis name 'x' or 'y'
    - **label** axis label such as 'Vdem'
    - **range_min** axis viewport minimum
    - **range_max** axis viewport maximum
    """
    IS_AXIS = False
    IS_LOG = False
    axis = None
    label = None
    units = None
    range_min = None
    range_max = None


    def setLog(self):
        """
        Set *IS_LOG = True* and convert range_(min/max) to log10
        """
        if not self.IS_LOG and self.range_min != None and self.range_max != None:
            self.setRange(numpy.log10(self.range_min), numpy.log10(self.range_max))
        self.IS_LOG = True

    def unsetLog(self):
        """
        Set *IS_LOG = False* and unlog range_(min/max)
        """
        if self.IS_LOG and self.range_min != None and self.range_max != None:
            self.setRange(10.0**self.range_min, 10.0**self.range_max)
        self.IS_LOG = False

    def setRange(self, min, max):
        """
        Set the axis range:

        - **min** lower bound
        - **max** upper bound
        """
        self.range_min = min
        self.range_max = max

    def setAxis(self, axis):
        """
        Enable this axis so that *IS_AXIS = True* where:

        - **axis** is either 'x' or 'y'

        If no *label* has been defined it is set to the axis name.
        """
        if axis in ['x','X']:
            self.IS_AXIS = True
            self.axis = 'x'
            if self.label == None or self.label in ['y','Y']:
                self.label = 'x'
        elif axis in ['y','Y']:
            self.IS_AXIS = True
            self.axis = 'y'
            if self.label == None or self.label in ['x','X']:
                self.label = 'y'
        else:
            print 'setAxis: %s is not a valid axis name' % axis

    def unsetAxis(self):
        """
        Disable this axis so that *IS_AXIS = False* and *axis = None*.
        """
        self.IS_AXIS = False
        self.axis = None


class GraphicsObj2(object):
    """
    Class describing a "graph" or collection of series.

    - **arr** an ndarray of data shape = (r, c>=1)
    - **labels** (optional) a list of column labels

    If labels are not supplied they are created as line1, line2 etc.
    Data series (LineObj2) objects are stored as GraphicsObj2
    instance attributes.
    """

    series = None
    scalar = None
    name = None

    def __init__(self, arr, labels=None):
        self.addSeriesFromArray(arr, labels)

    def setName(self, name):
        self.name = name

    def getName(self):
        return self.name

    def setAxis(self, axis, label):
        if hasattr(self, label):
            self.__getattribute__(label).axis.setAxis(axis)
        else:
            print 'setAxis: %s is not a valid label' % label

    def getAxis(self, axis, label=False):
        out = None
        for s in self.series:
            if s.axis.IS_AXIS and s.axis.axis == axis:
                if label:
                    out =  (s.getData(), s.getLabel())
                else:
                    out = s.getData()
                break
        return out

    def addOneSeries(self, series, label):
        L = LineObj2(series, label)
        self.series.append(L)
        self.__setattr__(label, L)

    def addSeriesFromArray(self, arr, labels=None):
        assert type(arr) == numpy.ndarray, '\nI need an ndarray'
        assert arr.shape[1] >= 1, '\nndarray must be 2D'
        self.series = []

        if labels == None or len(labels) != arr.shape[1]:
            labels = ['line%s' % c for c in range(arr.shape[1])]

        for c in range(arr.shape[1]):
            self.addOneSeries(arr.take([c], axis=1), labels[c])

    def hasSeries(self):
        return [l.getLabel() for l in self.series]

    def updateOneSeries(self, series, label):
        if hasattr(self, label):
            self.__getattribute__(label).setData(series)
        else:
            print '%s is an invalid series name.' % label

    def updateSeriesFromArray(self, arr):
        labels = self.hasSeries()
        if arr.shape[1] != len(labels):
            print 'updateSeriesFromArray: incorrect array shape for series (%s)' % self.hasSeries()
        else:
            for c in range(arr.shape[1]):
                self.updateOneSeries(arr.take([c], axis=1), labels[c])

    def delOneSeries(self, label):
        currS = self.hasSeries()
        if label in currS:
            self.series.pop(currS.index(label))
            self.__delattr__(label)

    def getAllSeriesAsArray(self):
        return numpy.hstack([s.getData() for s in self.series])

    def getAllSeriesAsDict(self):
        out = {}
        for s in self.series:
            out.update({s.getLabel() : s.getData()})
        return out

    def getAllSeriesAndProperties(self):
        out = {}
        for s in self.series:
            out.update({s.getLabel() : {'data' : s.getData(),
                                         'properties' : s.getProperties()}
                        })
        return out

    def setSeriesProperties(self, label, **kwargs):
        if hasattr(self, label):
            self.__getattribute__(label).setProperties(**kwargs)

    def setAllSeriesProperties(self, **kwargs):
        for s in self.hasSeries():
            self.setSeriesProperties(s, **kwargs)

    def unsetAllSeriesProperties(self):
        for s in self.hasSeries():
            self.__getattribute__(s).unsetAllProperties()
'''



"""
PySCeS - Python Simulator for Cellular Systems (http://pysces.sourceforge.net)

Copyright (C) 2004-2022 B.G. Olivier, J.M. Rohwer, J.-H.S Hofmeyr all rights reserved,

Brett G. Olivier (bgoli@users.sourceforge.net)
Triple-J Group for Molecular Cell Physiology
Stellenbosch University, South Africa.

Permission to use, modify, and distribute this software is given under the
terms of the PySceS (BSD style) license. See LICENSE.txt that came with
this distribution for specifics.

NO WARRANTY IS EXPRESSED OR IMPLIED.  USE AT YOUR OWN RISK.
Brett G. Olivier
"""

squidy =\
"""
     :vzW:.
        :9#@M8v:.    .#MMMMMMM#A;.             .:i,.
     :vv.    tMMMMMM.#MM@YivBMMMMMMMMMMMM@MMMM#Qbz1X
     ..,YEMQ,     7MMMMM#zi        ...  .ii..
  ...      vEMMMb@E:c$MMMMMMMz
   .i;7ZbIY:...UEBMMMMMMMMMMMM#                                ,MMMM.
         .:zZ#MMMZMMMMMMMMMMMMMb   iMMMMMMMMMMMMMMMMMMWt,     YMMMMMM#
      ....       .:c71tt80oM@@MMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMM@
        .ii7vt0MMMM09ZtnWMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMM,
       ...::cnWMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMUMMMM@MMMMMMMc
          .YUc    .;o@M@ZXEM0zB@MM@MMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMM6
 .YYtn9UIA$@6Z$W@E7:QMM0ZMMMMMMco; 1MMMMMMMMMMMMMMMMMMMMMMMMM#MMMMMMMM#
        ..      .t6MMMMMMMMMMMX            .t0@MMMMM@I,..      iMMMMM$
           ;QMMMMMWS1tQMMMMM@;                                   ibMi
     .ivE@M0v.  it@M@WMMM0c
  .,;Y:    .XE202i. .QM7
      .:Yt2c.      EMA
    ....           MM6  .:i;7z8Q#$EEz11Cv,.
                    #MMMMMMMMMMM@@#MMMMMMMM@@WEU7

Welcome to the Kraken distributed modelling environment
(c) Brett G. Olivier, Stellenbosch 2007-2017.
"""
octopussy =\
"""
                           .Yt90QW$EC.
                        .U@@@W8AzoI0MMW.
                       WMM@WEE88Z980W@MM@
                      MM$Q00bEbbbbEE0W$#MC
                     bMEbbbbbbbbb88bbE00@@
                     M$8bbZ8bb8bbbbbZZb0QM:
                     MBb8ZbbbbbbbbbbEbEEQ@U             ..
                     M@bZ8EbbbbbbbbbbbbEE##           tMMMM@Z;
                     7M8bbbbbbbbbbbbbbbbE@@          6M@#@@MMMMi
                     .Mbb888ZbZZ88Zb8Z88bMZ          iM@00#.
       ;M             @#88bZ86WQEWW8Z88b6M,            @B@M$
       MM.            UM88bb0#tY@,n@QZ8ZWM             ;MBMB
       MM@. .,i:      .ME8b8Q$It@28B8886Mi       .     EBUZ@7
i      MM@MMMMMM.      MQb8Z8600b08ZbbA@1       tMM@QEM@90#M@
@     MMMM@MM#$M0      z@98b8888888bE9WW        @M@@MM@$MMMM.
MS    Y2,.,,,@QBMMZ0MM#@@A8bEEbZQE6ZZb@#@M     .MWQ$W,...Y2.
#MI      ;St YM@@MM@#$00WZ8bbb8E@@69Z#E0#M@   .##bZM.         2A,   .0
E#MM;  6M@B$M    XM$#BE60bZbEb98QWZ9B06Q@@MMM#M#E6IM. .8MM#..MMMMQ;CQ#
:M#@MMMM$@#Q#B   7@Q0W#@0ZZZ9ZZZA9bbM0E$EEQ#MM#B@M0MBWMMB$MMMMBB#MMM@b
.MMMMM@#0.;MW@M#@$9E0Z9Z660W9b889666WB089bQ#WB@@#M@#@#$#@@#$QEE@MMM8,.
.@M6 .MMz  ,MM#BQ#@ZE@@BW##E88bb0$bUAU69b0MB622Q#7;QQb@MM1S@@@#7Q#.
       .    MM#@MI     iY0WAb#Q8EW##QE0EZ87i80EA    .   .   ..
                  Y21c. .EE8W.0#6AU#0. ;8b8ZQ#2.:U#@@@S
        c@@@c    $M@@@MM@09@i .Mb9EE@@$$QbEb690@M@W@M@MM
       .MM@@MM06#MQ$@#WQ0EZM.  8MM  .M@00#EMM@#B$@@MMMBMA
    :tAWME0#@@@@B8@MMMBIW@@B   MMQ    QQQ0 cMMUZS. :i @Mo
  .MMMMWZ6WZ.#M@W@W:E;                               :#i
  7MBW@@$@@  ZMZt9
Welcome to the Kraken distributed modelling environment
(c) Brett G. Olivier, Stellenbosch 2007-2017.
"""


"""
PySCeS - Python Simulator for Cellular Systems (http://pysces.sourceforge.net)

Copyright (C) 2004-2022 B.G. Olivier, J.M. Rohwer, J.-H.S Hofmeyr all rights reserved,

Brett G. Olivier (bgoli@users.sourceforge.net)
Triple-J Group for Molecular Cell Physiology
Stellenbosch University, South Africa.

Permission to use, modify, and distribute this software is given under the
terms of the PySceS (BSD style) license. See LICENSE.txt that came with
this distribution for specifics.

NO WARRANTY IS EXPRESSED OR IMPLIED.  USE AT YOUR OWN RISK.
Brett G. Olivier
"""
"""
METATOOL is a C program developed from 1998 to 2000 by Thomas Pfeiffer
(Berlin) in cooperation with Stefan Schuster and Ferdinand
Moldenhauer (Berlin) and Juan Carlos Nuno (Madrid).

See also
http://www.biologie.hu-berlin.de/biophysics/Theory/tpfeiffer/metatool.html.
"""

# Init file used for distutils install



import re
import os

cwd = os.path.dirname(__file__)

with open(os.path.join(cwd, 'version.txt'), 'r') as f:
    v = f.read().strip().split('.')

MAJOR = int(v[0])
MINOR = int(v[1])
MICRO = int(re.split('(\d+)', v[2])[1])
STATUS = re.split('(\d+)', v[2])[2]

def current_version():
    return '%s.%s.%s%s' % (MAJOR, MINOR, MICRO, STATUS)

def current_version_tuple():
    return (MAJOR, MINOR, MICRO)

__version__ = current_version()



"""
PySCeS - Python Simulator for Cellular Systems (http://pysces.sourceforge.net)

Copyright (C) 2004-2022 B.G. Olivier, J.M. Rohwer, J.-H.S Hofmeyr all rights reserved,

Brett G. Olivier (bgoli@users.sourceforge.net)
Triple-J Group for Molecular Cell Physiology
Stellenbosch University, South Africa.

Permission to use, modify, and distribute this software is given under the
terms of the PySceS (BSD style) license. See LICENSE.txt that came with
this distribution for specifics.

NO WARRANTY IS EXPRESSED OR IMPLIED.  USE AT YOUR OWN RISK.
Brett G. Olivier
"""

from Kraken import *
from KrakenDataTools import IntersectionAnalysis
os.chdir('c:/mypysces/btk')
##  os.chdir('/home/bgoli/mypysces/kraken')

START = time.time()
"""
Build controller and initialise active tentacles
Get a name for this task (task == collection of jobs)
"""
kcntrl = KrakenScanController()
kcntrl.setEnv(task_id='SCAN_A_P_A0512_big', working_dir='c:\\mypysces\\btk',\
        model_file='isola_fixA.psc', model_dir='c:\\mypysces\\pscmodels\\btk')
kcntrl.startController()
"""
Set up a list of 'static' initialisation commands that will be used to
initialise each server before a job is sent to it
"""
I_list = ['P_INIT,' + kcntrl.kc_model_server.model_file_name, 'P_LOAD',\
'P_SET_FLOAT,nleq2_nitmax,30',\
'P_SET_FLOAT,A05,1.2',\
'P_SET_FLOAT,V2,100.0', 'P_SET_FLOAT,V3,100.0']
kcntrl.setInitCmds(I_list)

pointsP1 = 14
pointsP2 = 1400
scan_segments = 512
##  pointsP1 = 20
##  pointsP2 = 100
##  scan_segments = 20
scan_job = 'P_SCAN, 2, 4, P, %s, %s, '+str(pointsP1)+', 1, 0, A, 0.01, 400,  '+str(pointsP2)+', 1, 0,  P, A, J_R1, J_R3'
kcntrl.setScanJobs(0.01, 40000, scan_segments, scan_job, log=True)

kcntrl.Run(raw_data_dump=False)
kcntrl.concatenateResults()
print('kcntrl.result_array.shape', kcntrl.result_array.shape)

MID = time.time()

kisect = IntersectionAnalysis()

kisect.zmin = 0.01
kisect.setCoords(kisect.DataFilter(kcntrl.getAndClearResultArray(), report=False))
##  kisect.setVTKCoords()

kisect.intersect_tol = 0.04
kisect.CalculateIntersect()
kisect.writeGnuPlot_3D(kisect.intersection, kcntrl.task_id+'_isect')
##  kcntrl.Dump(kisect.intersection, kcntrl.task_id+'_isect.bin')

##  print 'kisect.vtkcoords.shape, type', kisect.vtkcoords.shape, kisect.vtkcoords.dtype

##  kisect.LogVTKdata()
kisect.LogVTKIntersectiondata()
##  kcntrl.Dump(kisect.vtkcoords, kcntrl.task_id+'_vtkcoords.bin')

##  kisect.writeVTK_UnstructuredGrid(kisect.vtkcoords.take([0,1,2],axis=1), kcntrl.task_id+'_J1')
##  kisect.writeVTK_UnstructuredGrid(kisect.vtkcoords.take([0,1,3],axis=1), kcntrl.task_id+'_J3')
kisect.writeVTK_UnstructuredGrid(kisect.vtkintersection.take([0,1,2],axis=1), kcntrl.task_id+'_isect')

END = time.time()

print("\n********** start Kraken report **********")
print("Using %s tentacles" % len(kcntrl.kc_available_server_list))
print("Scan points          = %s" % (pointsP1*pointsP2*scan_segments))
print("Data generation time = %2.2f minutes." % ((MID-START)/60.0))
print("Data analysis time   = %2.2f minutes." % ((END-MID)/60.0))
print("Total scan time      = %2.2f minutes." % ((END-START)/60.0))
print("Total points/minute  = %2.1f" % ((60.0*pointsP1*pointsP2*scan_segments)/(END-START)))
print("********** end Kraken report **********\n")

kcntrl.Dump(kisect, kcntrl.task_id+'_kisect.bin')


"""
PySCeS - Python Simulator for Cellular Systems (http://pysces.sourceforge.net)

Copyright (C) 2004-2022 B.G. Olivier, J.M. Rohwer, J.-H.S Hofmeyr all rights reserved,

Brett G. Olivier (bgoli@users.sourceforge.net)
Triple-J Group for Molecular Cell Physiology
Stellenbosch University, South Africa.

Permission to use, modify, and distribute this software is given under the
terms of the PySceS (BSD style) license. See LICENSE.txt that came with
this distribution for specifics.

NO WARRANTY IS EXPRESSED OR IMPLIED.  USE AT YOUR OWN RISK.
Brett G. Olivier
"""

from __future__ import division, print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from .version import __version__

__doc__ = '''PySCeS ModelMap module: useful for exploring model component relations'''


class ModelMapBase(object):
    name = None

    def getName(self):
        return self.name

    def setName(self, name):
        self.name = name

    def get(self, attr):
        """Return an attribute whose name is str(attr)"""
        try:
            return getattr(self, attr)
        except:
            print("%s is not an attribute of this instance" % attr)
            return None


class MapList(list):
    def __init__(self, *args):
        list.__init__(self, *args)

    def asSet(self):
        return set(self.__getslice__(0, self.__len__()))


class ModelMap(ModelMapBase):
    __nDict__ = None
    reactions = None
    species = None
    species_fixed = None
    compartments = None
    __model__ = None
    __InitStrings__ = None
    __InitDict__ = None
    __not_inited__ = None
    global_parameters = None
    __parameter_store__ = None

    def __init__(self, model):
        self.setName(model.ModelFile[:-4])
        self.__nDict__ = model.__nDict__
        self.__model__ = model
        self.__InitDict__ = self.__model__.__InitDict__.copy()
        self.__compartments__ = self.__model__.__compartments__.copy()
        for k in list(self.__InitDict__.keys()):
            self.__InitDict__[k] = getattr(self.__model__, k)
        self.global_parameters = []
        self.__parameter_store__ = []
        self.__not_inited__ = []

        # operational shortcuts
        self.addSpecies()
        self.addReactions()
        self.generateMappings()
        self.addCompartments()

    def __cleanString__(self, s):
        s = s.lstrip()
        s = s.rstrip()
        return s

    def addOneSpecies(self, species, fix=False):
        s = Species(species)
        s.setValue(self.__InitDict__[s.name])
        if fix:
            s.fixed = True
        setattr(self, species, s)
        self.species.append(s)
        if fix:
            self.species_fixed.append(s)

    def addCompartments(self):
        self.compartments = []
        for c in self.__model__.__compartments__:
            co = Compartment(
                self.__model__.__compartments__[c]['name'],
                self.__model__.__compartments__[c]['size'],
            )
            self.compartments.append(co)
            setattr(self, c, co)
        cname = [c.name for c in self.compartments]
        for s in list(self.__model__.__sDict__.keys()):
            if self.__model__.__sDict__[s]['compartment'] in cname:
                getattr(self, self.__model__.__sDict__[s]['compartment']).setComponent(
                    getattr(self, s)
                )
                getattr(self, s).compartment = getattr(
                    self, self.__model__.__sDict__[s]['compartment']
                )
        for r in list(self.__model__.__nDict__.keys()):
            if self.__model__.__nDict__[r]['compartment'] in cname:
                getattr(self, self.__model__.__nDict__[r]['compartment']).setComponent(
                    getattr(self, r)
                )
                getattr(self, r).compartment = getattr(
                    self, self.__model__.__nDict__[r]['compartment']
                )

    def addOneReaction(self, reaction):
        r = Reaction(reaction)
        r.addFormula(self.__nDict__[r.name]['RateEq'].replace('self.', ''))
        if self.__nDict__[r.name]['Type'] == 'Irrev':
            r.reversible = False

        fxnames = self.hasFixedSpecies()
        for p in self.__nDict__[r.name]['Params']:
            p = p.replace('self.', '')
            if (
                p not in self.hasGlobalParameters()
                and p not in fxnames
                and p not in self.__compartments__
            ):
                if p in self.__InitDict__:
                    par = Parameter(p, self.__InitDict__[p])
                else:
                    par = Parameter(p)
                    if p not in self.__not_inited__:
                        self.__not_inited__.append(p)
                par.setAssociation(r)
                self.global_parameters.append(par)
                setattr(self, p, par)
                r.addParameter(par)
            elif p not in fxnames and p not in self.__compartments__:
                pidx = self.hasGlobalParameters().index(p)
                self.global_parameters[pidx].setAssociation(r)
                r.addParameter(self.global_parameters[pidx])
        setattr(self, reaction, r)
        self.reactions.append(r)

    def addSpecies(self):
        self.species = []
        self.species_fixed = []
        for s in self.__model__.species:
            self.addOneSpecies(s, fix=False)
        for s in self.__model__.fixed_species:
            self.addOneSpecies(s, fix=True)

    def addReactions(self):
        self.reactions = []
        for r in self.__model__.reactions:
            self.addOneReaction(r)

    def generateMappings(self):
        for reac in self.reactions:
            for reag in self.__nDict__[reac.name]['Reagents']:
                if self.__nDict__[reac.name]['Reagents'][reag] < 0.0:
                    reac.addSubstrate(getattr(self, reag.replace('self.', '')))
                    getattr(self, reag.replace('self.', '')).setSubstrate(
                        getattr(self, reac.name)
                    )
                else:
                    reac.addProduct(getattr(self, reag.replace('self.', '')))
                    getattr(self, reag.replace('self.', '')).setProduct(
                        getattr(self, reac.name)
                    )
                reac.stoichiometry.setdefault(
                    reag.replace('self.', ''),
                    self.__nDict__[reac.name]['Reagents'][reag],
                )
            for mod in self.__nDict__[reac.name]['Modifiers']:
                reac.addModifier(getattr(self, mod.replace('self.', '')))
                getattr(self, mod.replace('self.', '')).setModifier(
                    getattr(self, reac.name)
                )

    def hasReactions(self):
        return MapList([r.name for r in self.reactions])

    def hasSpecies(self):
        return MapList([s.name for s in self.species])

    def hasFixedSpecies(self):
        return MapList([s.name for s in self.species_fixed])

    def findReactionsThatIncludeAllSpecifiedReagents(self, *args):
        assert len(args) > 1, '\nNeed two or more species for this one!'
        setlist = [getattr(self, s).isReagentOf().asSet() for s in args]
        isect = setlist[0]
        for s in setlist:
            isect.intersection_update(s)
        return MapList(isect)

    def hasGlobalParameters(self):
        return MapList(p.name for p in self.global_parameters)


class Reaction(ModelMapBase):
    modifiers = None
    substrates = None
    products = None
    stoichiometry = None
    parameters = None
    reversible = True
    formula = None
    compartment = None

    def __init__(self, name):
        self.setName(name)
        self.modifiers = []
        self.substrates = []
        self.products = []
        self.stoichiometry = {}
        self.parameters = []

    def addSubstrate(self, species):
        setattr(self, species.name, species)
        self.substrates.append(species)

    def addProduct(self, species):
        setattr(self, species.name, species)
        self.products.append(species)

    def addModifier(self, species):
        setattr(self, species.name, species)
        self.modifiers.append(species)

    def addFormula(self, formula):
        self.formula = formula

    def addParameter(self, par):
        setattr(self, par.name, par)
        self.parameters.append(par)

    def hasProducts(self, t=type):
        return MapList([p.name for p in self.products])

    def hasSubstrates(self):
        return MapList([s.name for s in self.substrates])

    def hasModifiers(self):
        return MapList([m.name for m in self.modifiers])

    def hasParameters(self):
        return MapList([p.name for p in self.parameters])

    def hasReagents(self):
        return MapList(self.hasSubstrates() + self.hasProducts())


class NumberBase(ModelMapBase):
    value = None

    def __call__(self):
        return self.value

    def getValue(self):
        return self.value

    def setValue(self, v):
        self.value = v


class Species(NumberBase):
    subs = None
    prods = None
    mods = None
    fixed = False
    compartment = None

    def __init__(self, name):
        self.setName(name)
        self.subs = []
        self.prods = []
        self.mods = []

    def setSubstrate(self, reaction):
        setattr(self, reaction.name, reaction)
        self.subs.append(reaction)

    def setProduct(self, reaction):
        setattr(self, reaction.name, reaction)
        self.prods.append(reaction)

    def setModifier(self, reaction):
        setattr(self, reaction.name, reaction)
        self.mods.append(reaction)

    def isSubstrateOf(self):
        return MapList([r.name for r in self.subs])

    def isProductOf(self):
        return MapList([r.name for r in self.prods])

    def isModifierOf(self):
        return MapList([r.name for r in self.mods])

    def isReagentOf(self):
        return MapList(self.isSubstrateOf() + self.isProductOf())


class Parameter(NumberBase):
    association = None
    formula = None

    def __init__(self, name, value=None):
        self.name = name
        self.value = value
        self.association = []

    def setAssociation(self, reac):
        self.association.append(reac)
        setattr(self, reac.name, reac)

    def isParameterOf(self):
        return MapList([a.name for a in self.association])

    def setFormula(self, formula):
        self.formula = formula


class Compartment(NumberBase):
    components = None

    def __init__(self, name, value=None):
        self.name = name
        self.value = value
        self.components = []

    def setComponent(self, comp):
        self.components.append(comp)
        setattr(self, comp.name, comp)

    def hasComponents(self):
        return MapList([a.name for a in self.components])


if __name__ == '__main__':
    import pysces

    M = pysces.model('pysces_model_linear1')
    M.doLoad()

    print('\nModel', M.ModelFile)
    print('=============')
    modmap = ModelMap(M)

    print('Reactions\n', modmap.hasReactions())
    print('Species\n', modmap.hasSpecies())
    print('FixedSpecies\n', modmap.hasFixedSpecies())
    print(' ')
    print('R1 has reagents\n', modmap.R1.hasReagents())
    print('R1 has sub\n', modmap.R1.hasSubstrates())
    print('R1 has prod\n', modmap.R1.hasProducts())
    print('R1 has mod\n', modmap.R1.hasModifiers())
    print(' ')
    print('s2 is reagent\n', modmap.s2.isReagentOf())
    print('s2 is sub\n', modmap.s2.isSubstrateOf())
    print('s2 is prod\n', modmap.s2.isProductOf())
    print('s2 is mod\n', modmap.s2.isModifierOf())
    print(' ')
    print('R2 stoich\n', modmap.R2.stoichiometry)
    print(' ')
    print(
        'findReactionsThatIncludeAllSpecifiedReagents(A, B):',
        modmap.findReactionsThatIncludeAllSpecifiedReagents('s1', 's2'),
    )

    print('\nmodmap.hasGlobalParameters\n', modmap.hasGlobalParameters())

    print('\nParameter associations')
    for p in modmap.global_parameters:
        print('%s.isParameterOf() %s' % (p.name, p.isParameterOf()))
